/*
 * Copyright (c) 2017-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <queue>
#include <unistd.h>

#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibdiag/ibdiag_types.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibdm/SysLogger.h>

struct NodeDirectRoute {
    IBNode* node;
    list_p_direct_route direct_routes;
};

typedef queue < NodeDirectRoute* > queue_p_node_direct_route;

static NodeDirectRoute* pop_node(queue_p_node_direct_route& queue)
{
    NodeDirectRoute* node = queue.front();
    queue.pop();
    return node;
}

static direct_route_t* createNextDR(direct_route_t* cur_dr, u_int8_t port)
{
    direct_route_t* next_dr = new direct_route_t;
    memcpy(next_dr, cur_dr, sizeof(direct_route_t));
    next_dr->path.BYTE[cur_dr->length] = port;
    next_dr->length = (u_int8_t)(cur_dr->length + 1);

    IBDIAG_RETURN(next_dr);
}

inline void IBDiag::AddGoodPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddGoodPath %s\n",
              Ibis::ConvertDirPathToStr(p_direct_route).c_str());

    this->good_direct_routes.push_back(p_direct_route);
}

static void addSwitchToQueue(map_guid_list_p_direct_route& nodes_map,
                             map_guid_list_p_direct_route& ports_map,
                             NodeDirectRoute* cur_switch,
                             queue_p_node_direct_route& bfs_queue)
{
    // add the switch to bfs_known_node_guids
    nodes_map[cur_switch->node->guid_get()] = cur_switch->direct_routes;
    // for switches port 0 guid is the same as node guid.
    ports_map[cur_switch->node->guid_get()] = cur_switch->direct_routes;

    bfs_queue.push(cur_switch);
}

int IBDiag::BuildDirectRoutesDB()
{
    local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS];
    u_int32_t local_ports_num;
    queue_p_node_direct_route bfs_queue;
    NodeDirectRoute* bfs_cur_node = new NodeDirectRoute;
    list_p_direct_route::iterator iter;
    IBPort* root_port = NULL;


    // get the root node
    if (this->GetAllLocalPortGUIDs(local_ports_array, &local_ports_num))
    {
        delete bfs_cur_node;
        IBDIAG_RETURN(1);
    }

    for (u_int32_t i = 0; i < local_ports_num; ++i) {

        if (local_ports_array[i].logical_state <= IB_PORT_STATE_DOWN)
            continue;

        u_int64_t port_guid = ADB2C_BE64_TO_CPU(local_ports_array[i].guid);
        map_guid_pport::iterator it = discovered_fabric.PortByGuid.find(port_guid);

        if (it != discovered_fabric.PortByGuid.end() && !this->ibis_obj.CheckDevicePort(
                local_ports_array[i].ca_name, (phys_port_t)local_ports_array[i].portnum)) {

            root_port = it->second;
            bfs_cur_node->node = root_port->p_node;
            this->root_node = bfs_cur_node->node;
            this->root_port_num = root_port->num;

            break;
        }
    }

    if (!root_port) {
        this->SetLastError("Load from file failed, err=Local InfiniBand ports not found in " DB_FILE);
        delete bfs_cur_node;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }

    // set device name and port for ibis
    if (this->SetPort(root_port->guid_get())) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        delete bfs_cur_node;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }

    INFO_PRINT("Using local IB device %s:%d\n", this->ibis_obj.GetDevName().c_str(),
               this->ibis_obj.GetPortNum());
    SysLogger::instance().syslog("Using local IB device %s:%d\n",
                                 this->ibis_obj.GetDevName().c_str(), this->ibis_obj.GetPortNum());

    direct_route_t* cur_dr = new direct_route_t;
    CLEAR_STRUCT(*cur_dr);
    cur_dr->length = 1;
    bfs_cur_node->direct_routes.push_back(cur_dr);

    if (bfs_cur_node->node->type == IB_SW_NODE) {

        addSwitchToQueue(bfs_known_node_guids, bfs_known_port_guids, bfs_cur_node, bfs_queue);
        this->SetDR(root_port, cur_dr);

    } else { // root is a CA, push the switch connected to it to the queue

        // add the CA to bfs_known
        bfs_known_node_guids[bfs_cur_node->node->guid_get()].push_back(cur_dr);
        bfs_known_port_guids[root_port->guid_get()].push_back(cur_dr);
        this->SetDR(root_port, cur_dr);

        this->AddGoodPath(cur_dr);

        // create the switch
        bfs_cur_node->node = root_port->p_remotePort->p_node;
        bfs_cur_node->direct_routes.clear();

        direct_route_t* sw_dr = createNextDR(cur_dr, (u_int8_t)root_port->num);
        bfs_cur_node->direct_routes.push_back(sw_dr);

        this->SetDR(root_port->p_remotePort, sw_dr);
        addSwitchToQueue(bfs_known_node_guids, bfs_known_port_guids, bfs_cur_node, bfs_queue);
    }

    while (!bfs_queue.empty()) {
        bfs_cur_node = pop_node(bfs_queue);

        for (phys_port_t i = 0; i <= bfs_cur_node->node->numPorts; ++i) { // go over all ports of the switch
            IBPort* port = bfs_cur_node->node->getPort(i);

            if (!port || !port->p_remotePort) {
                continue;
            }

            IBNode* remote_node = port->p_remotePort->p_node;

            if (remote_node->type == IB_SW_NODE) {  // switch, push to the bfs queue
                if (bfs_known_node_guids.find(remote_node->guid_get()) ==
                        bfs_known_node_guids.end()) { // not visited node

                    NodeDirectRoute* bfs_remote_node = new NodeDirectRoute;
                    bfs_remote_node->node = remote_node;

                    for (iter = bfs_cur_node->direct_routes.begin();
                         iter != bfs_cur_node->direct_routes.end(); ++iter) {

                        cur_dr = createNextDR(*iter, i);
                        bfs_remote_node->direct_routes.push_back(cur_dr);
                        this->AddGoodPath(cur_dr);
                    }

                    if (!bfs_remote_node->direct_routes.empty())
                        this->SetDR(port->p_remotePort,
                                                  bfs_remote_node->direct_routes.front());

                    addSwitchToQueue(bfs_known_node_guids, bfs_known_port_guids,
                                     bfs_remote_node, bfs_queue);

                }
            } else { // CA, push to bfs_known

                for (iter = bfs_cur_node->direct_routes.begin();
                     iter != bfs_cur_node->direct_routes.end(); ++iter) {

                    cur_dr = createNextDR(*iter, i);
                    bfs_known_node_guids[remote_node->guid_get()].push_back(cur_dr);
                    bfs_known_port_guids[port->p_remotePort->guid_get()].push_back(cur_dr);
                    this->SetDR(port->p_remotePort, cur_dr);
                    this->AddGoodPath(cur_dr);
                }
            }
        }
        delete bfs_cur_node;
    }
    this->ibdiag_discovery_status = DISCOVERY_SUCCESS;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

