/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdiag/ibdiag_flid.h"

using namespace std;

bool FLIDsManager::IsApplicable() const
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN (!this->globalRanges.empty() || !this->localRanges.empty());
}

int FLIDsManager::CheckLocalSubnet(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = this->CheckRoutersRanges(errorsList);
    if (rc)
        IBDIAG_RETURN(rc);

    if (!this->IsApplicable())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->CheckLocalAndGlobalRangesCorrectness(errorsList);

    IBDIAG_RETURN(rc);
}

int FLIDsManager::CheckRoutersRanges(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    for (set_pnode::const_iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->Routers.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->Routers.end(); ++nI) {
        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->error = "DB error - found null node in Routers set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        const SMP_RouterInfo* p_router_info =
            this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterInfo(p_curr_node->createIndex);
        if (!p_router_info)
            continue;

        if (!this->IsConfiguredFLID(*p_curr_node, *p_router_info))
            continue;

        this->localRanges[range_t(p_router_info->local_router_lid_base,
                                    p_router_info->local_router_lid_top)].push_back(p_curr_node);
        this->globalRanges[range_t(p_router_info->global_router_lid_base,
                                    p_router_info->global_router_lid_top)].push_back(p_curr_node);
    }

    int rc = this->CheckRanges(this->globalRanges, errorsList, true);
    if (rc)
        IBDIAG_RETURN(rc);

    if (!this->globalRanges.empty()) {
        //we take the first range even if there are some of them
        this->p_ibdiag->GetDiscoverFabricPtr()->globalMinFLID =
                    static_cast<lid_t>(globalRanges.begin()->first.start);
        this->p_ibdiag->GetDiscoverFabricPtr()->globalMaxFLID =
                    static_cast<lid_t>(globalRanges.begin()->first.end);
    }

    rc = this->CheckRanges(this->localRanges, errorsList, false);

    IBDIAG_RETURN(rc);
}

bool FLIDsManager::IsConfiguredFLID(const IBNode &routerNode, const SMP_RouterInfo &routerInfo) const
{
    IBDIAGNET_ENTER;

    if (!this->p_ibdiag->GetCapabilityModulePtr()->IsSupportedSMPCapability(&routerNode,
                                                                    EnSMPCapIsRouterLIDSupported))
        IBDIAG_RETURN(false);

    if (!routerInfo.global_router_lid_base     &&
            !routerInfo.global_router_lid_top  &&
            !routerInfo.local_router_lid_base  &&
            !routerInfo.local_router_lid_top)
        IBDIAG_RETURN(false);

    IBDIAG_RETURN(true);
}

int FLIDsManager::CheckRanges(const ranges_map &ranges,
                            list_p_fabric_general_err &errorsList, bool isGlobal)
{
    IBDIAGNET_ENTER;

    if (ranges.empty())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (ranges.size() == 1) {
        INFO_PRINT("All routers in the subnet have the same %s FLID range: start=%d end=%d\n",
                    (isGlobal ? "global" : "local"),
                    ranges.begin()->first.start, ranges.begin()->first.end);
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    stringstream ss;
    ss << "Different "
       << (isGlobal ? "global " : "local ")
       << "FLID ranges were detected for the routers: ";
    int rc = this->RangesToStream(ranges, ss, 3);
    if (rc)
        IBDIAG_RETURN(rc);

    errorsList.push_back(new FLIDError(ss.str()));

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::RangesToStream(const ranges_map &ranges, ostream &stream, int inLineNum)
{
    IBDIAGNET_ENTER;

    for (ranges_map::const_iterator itr = ranges.begin(); itr != ranges.end(); itr++) {
        int rc = this->GUIDsToStream(itr->second, stream, inLineNum);
        if (rc)
            IBDIAG_RETURN(rc);

        stream << "(total " << itr->second.size() << "),"
               << " start=" << itr->first.start << " end=" << itr->first.end
               << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::CheckLocalAndGlobalRangesCorrectness(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    if (this->globalRanges.size() != 1 || this->localRanges.size() != 1) {
        //Do we need any warning message here?
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    if (!this->localRanges.begin()->first.end && !this->localRanges.begin()->first.start) {
        INFO_PRINT("The subnet does not have its own FLIDs\n");
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    //check local range is in the global one
    if (this->globalRanges.begin()->first.start > this->localRanges.begin()->first.start ||
            this->globalRanges.begin()->first.end < this->localRanges.begin()->first.end) {
        stringstream ss;
        ss  << "Local FLID range is not in the global one." << endl
            << " Local range: start=" << this->localRanges.begin()->first.start
            << " end= " << this->localRanges.begin()->first.end << endl
            << " global range: start=" << this->globalRanges.begin()->first.start
            << " end=" << this->globalRanges.begin()->first.end;

        errorsList.push_back(new FLIDError(ss.str()));
    } else
        INFO_PRINT("Local FLID range is in the global one\n");

    //Local subnet LIDs and Global FLIDs sets should not intersect
    this->FindCommonLids();
    if (!this->common.empty()) {
        stringstream ss;
        ss << "Local LIDs";
        this->LidsToStream(this->common, ss, 3);
        ss << " found in Global FLIDs range "
           << '(' << this->globalRanges.begin()->first.start
           << " ... "  << this->globalRanges.begin()->first.end << ')'
           << endl;

        errorsList.push_back(new FLIDError(ss.str()));
    }
    else
       INFO_PRINT("Local subnet LID and global FLID ranges are OK\n");

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//local subnet lids should not be a continues single range.
//it could be like this:
//[local subnet lids] [global FLIDs] [continued local subnet lids]
//all laying in the UNICAST lid range.
void FLIDsManager::FindCommonLids()
{
    IBDIAGNET_ENTER;

    IBFabric *p_fabric = this->p_ibdiag->GetDiscoverFabricPtr();;
    //here we are sure globalRanges has only one record
    range_t range = this->globalRanges.begin()->first;
    common.clear();

    for (lid_t lid = p_fabric->minLid; lid <= p_fabric->maxLid; lid++) {
        if (!p_fabric->getPortByLid(lid))
            continue;

        if (range.start <= lid && lid <= range.end)
            common.push_back(lid);
    }

    IBDIAG_RETURN_VOID;
}

bool FLIDsManager::FindIntersection(lid_t min1, lid_t max1, lid_t min2, lid_t max2,
                                    std::pair<lid_t, lid_t> &common) const
{
    IBDIAGNET_ENTER;

    common.first = common.second = 0;

    if (min1 <= min2 && min2 <= max1) {
        common.first = min2;
        common.second = (max2 <= max1) ? max2 : max1;

        IBDIAG_RETURN(true);
    } else if (min1 <= max2 && max2 <= max1) {
        common.first = (min2 <= min1) ? min1 : min2;
        common.second = max2;

        IBDIAG_RETURN(true);
    }

    IBDIAG_RETURN(false);
}

bool FLIDsManager::FindIntersection(const ranges_map &ranges_1, const ranges_map &ranges_2,
                          std::pair<lid_t, lid_t> &common) const
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(this->FindIntersection(static_cast<lid_t>(ranges_1.begin()->first.start),
                                         static_cast<lid_t>(ranges_1.begin()->first.end),
                                         static_cast<lid_t>(ranges_2.begin()->first.start),
                                         static_cast<lid_t>(ranges_2.begin()->first.end),
                                         common));
}

int FLIDsManager::CheckAdjacentSubnets(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = this->CollectAdjSubnetsRanges(errorsList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = this->CheckAdjSubnetsRanges(errorsList);

    IBDIAG_RETURN(rc);
}

int FLIDsManager::CollectAdjSubnetsRanges(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    const IBFabric &discoverdFabric =  *this->p_ibdiag->GetDiscoverFabricPtr();
    for (set_pnode::const_iterator nI = discoverdFabric.Routers.begin();
            nI != discoverdFabric.Routers.end(); ++nI) {
        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->error = "DB error - found null node in Routers set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        const SMP_RouterInfo* p_router_info =
            this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterInfo(p_curr_node->createIndex);
        if (!p_router_info)
            continue;

        if (!this->IsConfiguredFLID(*p_curr_node, *p_router_info))
            continue;


        const size_t size =  ARRAY_SIZE(SMP_AdjSubnetsRouterLIDInfoTable::Record);
        u_int8_t blockNum = 0;
        u_int8_t idx = 0;
        const SMP_AdjSubnetsRouterLIDInfoTable *p_adjSubnetTable = NULL;
        for (u_int8_t j = 0; j < p_router_info->AdjacentSubnetsRouterLIDInfo; j++) {
                // get the next block, initialized already to first
                if ((j % size) == 0) {
                        blockNum = static_cast<u_int8_t>(j / size);
                        p_adjSubnetTable =
                         this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPAdjSubnteRouterLIDInfoTbl(
                                                                        p_curr_node->createIndex,
                                                                        blockNum);
                }

                if (!p_adjSubnetTable)
                    continue;

                idx = static_cast<u_int8_t>(j % size);
                const AdjSubnetRouterLIDRecord &record = p_adjSubnetTable->Record[idx];
                if (!record.subnet_prefix_id) {
                    stringstream ss;
                    ss << "Invalid subnet prefix ID=0 was detected on the router:"
                       << " name=" << p_curr_node->getName()
                       << " GUID=" << PTR(p_curr_node->guid_get())
                       << " record=" << +idx;

                    errorsList.push_back(new FLIDError(ss.str()));
                    continue;
                }

                ranges_map &rangesMap = this->adjSubnetsRanges[record.subnet_prefix_id];
                rangesMap[this->MakeRange(record)].push_back(p_curr_node);
        }//for on table indexes
    }//for on Nodes

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

FLIDsManager::range_t FLIDsManager::MakeRange(const AdjSubnetRouterLIDRecord &record) const
{
    IBDIAGNET_ENTER;

    range_t ret;
    ret.start = (record.local_router_lid_start << 4) | (0xF & record.local_router_lid_start_cont);
    ret.end = record.local_router_lid_end;

    IBDIAG_RETURN(ret);
}

int FLIDsManager::FindLocalSubnetPrefix()
{
    IBDIAGNET_ENTER;

    bool found = false;
    for (set_pnode::const_iterator
            nI = this->p_ibdiag->GetDiscoverFabricPtr()->HCAs.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->HCAs.end() && !found; ++nI) {
        const IBNode *p_node = *nI;
        if (!p_node) {
            this->error = "DB error - found null node in HCAs set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (p_node->isSpecialNode())
            continue;

        for (phys_port_t i = 0 ; i <=  p_node->numPorts; i++) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port                      ||
                    !p_port->p_remotePort    ||
                    !p_port->is_data_worthy())
                continue;

            const struct SMP_PortInfo *p_curr_port_info =
                    this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);
            if (!p_curr_port_info)      //port info not exist
                continue;

            //layout: [10 bits][38 bits][16 bits subnet prefix]
            this->localSubnetPrefix = 0xFFFF & p_curr_port_info->GIDPrfx;
            found = true;
            break;
        }
    }

    if (!found) {
        this->error = "Failed to collect local subnet prefix";
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::CheckAdjSubnetsRanges(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = FindLocalSubnetPrefix();
    if (rc)
        IBDIAG_RETURN(rc);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Local subnet prefix: " U16H_FMT "\n", this->localSubnetPrefix);

    //the same range is defined on all routers
    //and range is in the Global FLID range
    for (adj_subnets_ranges_map::const_iterator itr = this->adjSubnetsRanges.begin();
            itr != this->adjSubnetsRanges.end(); itr++) {
        if (this->localSubnetPrefix == itr->first)
            continue;

        const ranges_map  &ranges = itr->second;
        if (ranges.size() != 1) {
            stringstream ss;
            ss << "Different "
               << "FLID ranges were detected for the subnet: "
               << PTR(itr->first) << " on routers:"
               << endl;
            int rc = this->RangesToStream(ranges, ss, 3);
            if (rc)
                IBDIAGNET_RETURN(rc);

            errorsList.push_back(new FLIDError(ss.str()));
            continue;
        }

        INFO_PRINT("All routers in the subnet: " U16H_FMT
                " have the same FLID range: start=%d end=%d\n",
                itr->first,
                ranges.begin()->first.start, ranges.begin()->first.end);

        if (this->globalRanges.size() != 1)
            continue;

        if (this->globalRanges.begin()->first.start > itr->second.begin()->first.start ||
                this->globalRanges.begin()->first.end < itr->second.begin()->first.end) {
            stringstream ss;
            ss  << "Subnet: " << PTR(itr->first)
                << " local FLID range is not in the global one." << endl
                << " Local range: start=" << itr->second.begin()->first.start
                << " end= " << itr->second.begin()->first.end << endl
                << " global range: start=" << this->globalRanges.begin()->first.start
                << " end=" << this->globalRanges.begin()->first.end;

            errorsList.push_back(new FLIDError(ss.str()));
        }
    }//end for

    //ranges in the subnets do not intersect
    PairsContainer<u_int16_t> checkedPairs;
    for (adj_subnets_ranges_map::const_iterator itr1 = this->adjSubnetsRanges.begin();
            itr1 != this->adjSubnetsRanges.end(); itr1++) {
        for (adj_subnets_ranges_map::const_iterator itr2 = this->adjSubnetsRanges.begin();
            itr2 != this->adjSubnetsRanges.end(); itr2++) {

            //same subnet prefixes
            if (itr1->first == itr2->first || checkedPairs.Contains(itr1->first, itr2->first))
                continue;

            checkedPairs.Add(itr1->first, itr2->first);

            const ranges_map & ranges1 = itr1->second;
            const ranges_map & ranges2 = itr2->second;

            //different ranges were found
            if (ranges1.size() != 1 || ranges2.size() != 1) {
                continue;
            }

            std::pair<lid_t, lid_t> common;
            if (this->FindIntersection(ranges1, ranges2, common) ||
                    this->FindIntersection(ranges2, ranges1, common)) {
                stringstream ss;
                ss  << "Overlapping FLID ranges in the subnets: " << endl
                    << "    " << PTR(itr1->first) << ", ["
                    << ranges1.begin()->first.start << " ... " << ranges1.begin()->first.end
                    << ']' << endl
                    << "    " << PTR(itr2->first) << ", ["
                    << ranges2.begin()->first.start << " ... " << ranges2.begin()->first.end
                    << ']';

                errorsList.push_back(new FLIDError(ss.str()));
            }
        }//end for
    }//end for

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::CheckHCAsAndSwitches(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = this->CheckHCAs(errorsList);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->CheckSwitches(errorsList);

    IBDIAGNET_RETURN(rc);
}

int FLIDsManager::CheckHCAs(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    set_const_p_node reportedNodes;

    for (set_pnode::const_iterator itr = this->p_ibdiag->GetDiscoverFabricPtr()->Switches.begin();
            itr != this->p_ibdiag->GetDiscoverFabricPtr()->Switches.end(); itr++) {
        const IBNode* p_switch = *itr;
        if (!p_switch) {
            this->error = "DB error - found null node in Switches set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        set_lid flids;
        lid_to_ports_map hcaPortsByFLID;
        for (phys_port_t i = 1 ; i <=  p_switch->numPorts ; i++) {
            const IBPort *p_currPort = p_switch->getPort(i);
            if (!p_currPort) //port not exist
                continue;

            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_currPort->is_data_worthy())
                continue;

            if (!p_currPort->p_remotePort || !p_currPort->p_remotePort->p_node)
                continue;

            const IBNode *p_remoteNode = p_currPort->p_remotePort->p_node;
            if (p_remoteNode->type != IB_CA_NODE || p_remoteNode->isSpecialNode())
                continue;

            if (!this->p_ibdiag->GetCapabilityModulePtr()->IsSupportedSMPCapability(p_remoteNode,
                                                                EnSMPCapIsRouterLIDSupported))
                continue;

            lid_t flid = 0;
            if (!this->GetFLID(*(p_currPort->p_remotePort), flid))
                continue;

            if (!flid) {
                //FLID=0 can be configured if network does not have its own FLIDs
                if (this->localRanges.size() != 1 ||
                        this->localRanges.begin()->first.start || this->localRanges.begin()->first.end)
                    errorsList.push_back(new FLIDPortError(p_currPort->p_remotePort,
                                                    "FLID=0 is configured on the port"));
                continue;
            }

            if (this->localRanges.size() == 1) {
                if (this->localRanges.begin()->first.start || this->localRanges.begin()->first.end) {
                    this->CheckRouterLIDEnablementBitOnHCA(reportedNodes, *p_remoteNode,
                                    *p_currPort->p_remotePort, errorsList);

                    if (this->localRanges.begin()->first.start > flid ||
                            flid > this->localRanges.begin()->first.end) {
                        stringstream ss;
                        ss << "FLID=" << flid << " assigned to the port"
                           << " does not belong to the local FLIDs range "
                           << '(' << this->localRanges.begin()->first.start
                           << " ... " << this->localRanges.begin()->first.end << ')'
                           << endl;

                        errorsList.push_back(
                            new FLIDPortError(p_currPort->p_remotePort, ss.str()));
                    }
                } else {
                    stringstream ss;
                    ss << "FLID=" << flid << " assigned to the port"
                        << " but the network does not have its own FLIDs "
                        << '(' << this->localRanges.begin()->first.start
                        << " ... " << this->localRanges.begin()->first.end << ')'
                        << endl;

                    errorsList.push_back(new FLIDPortError(p_currPort->p_remotePort, ss.str()));
                }
            }

            flids.insert(flid);
            hcaPortsByFLID[flid].push_back(p_currPort->p_remotePort);
        }

        if (flids.empty()) {
            //think of message
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "None of FLIDS is associated with the switch GUID= "
                    U64H_FMT "\n", p_switch->guid_get());
           continue;
        }

        if (flids.size() != 1) {
            stringstream ss;
            ss << "Different FLIDS found for switch name=" << p_switch->getName()
               << " GUID=" << PTR(p_switch->guid_get()) << endl;
            int rc = this->FLIDsToStream(hcaPortsByFLID, ss, 3);
            if (rc)
                IBDIAG_RETURN(rc);

            errorsList.push_back(new FLIDError(ss.str()));

            this->swicthesWithMultipleFLIDs[p_switch].swap(hcaPortsByFLID);
        }
        else
            this->swicthesByFLID[*(flids.begin())].push_back(p_switch);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FLIDsManager::CheckRouterLIDEnablementBitOnHCA(set_const_p_node &reportedNodes,
                                          const IBNode &remoteNode,
                                          const IBPort &remotePort,
                                          list_p_fabric_general_err &errorsList) const
{
    IBDIAGNET_ENTER;

    const std::pair<set_const_p_node::iterator, bool> ret = reportedNodes.insert(&remoteNode);
    if (ret.second) {
        if (!this->p_ibdiag->GetCapabilityModulePtr()->IsSupportedSMPCapability(&remoteNode,
                                                        EnSMPCapIsExtendedPortInfoSupported)) {
            errorsList.push_back(new FLIDNodeError(&remoteNode,
                                                   "ExtenedPortInfo is not supported on the node. "
                                                   "RouterLIDEn bit cannot be enabled on any port"));
            IBDIAG_RETURN_VOID;
        }
    }

    const SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
                this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPMlnxExtPortInfo(
                                                                    remotePort.createIndex);
    //The bit RouterLIDEn indicates an ability to route by FLIDs on sending packets
    //back to the source. If it is disabled a returning packet will NOT route by FLIDs.
    //But packets to the destination still can be sent with FLIDs.
    if (p_curr_mlnx_ext_port_info && !p_curr_mlnx_ext_port_info->RouterLIDEn)
        errorsList.push_back(new FLIDPortError(&remotePort,
                                               "RouterLIDEn bit is not enabled on the port"));

    IBDIAG_RETURN_VOID;
}

int FLIDsManager::FLIDsToStream(const lid_to_ports_map &hcaPortsByFLID, ostream &stream, int inLineNum)
{
    IBDIAGNET_ENTER;

    for (lid_to_ports_map::const_iterator itr = hcaPortsByFLID.begin(); itr != hcaPortsByFLID.end(); itr++) {
        stream << "  FLID=" << itr->first << " CA ports(total " << itr->second.size() << "):";
        int rc = this->GUIDsToStream(itr->second, stream, inLineNum);
        if (rc)
            IBDIAG_RETURN(rc);

        stream << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FLIDsManager::LidsToStream(const vec_lids &lids, ostream &stream, int inLineNum) const
{
    IBDIAGNET_ENTER;

    //just to be on the safe side.
    if (lids.empty()) {
       stream << "[]";
       IBDIAG_RETURN_VOID;
    }

    stream << '[';
    vec_lids::const_iterator itrLast = --lids.end();

    int j = 0;
    int inLine = (inLineNum > 0) ? inLineNum : static_cast<int>(lids.size());
    for (vec_lids::const_iterator itr = lids.begin(); itr != itrLast && j < inLine; itr++) {
        stream << *itr << ", ";
        j++;
    }

    if (j + 1 == static_cast<int>(lids.size()) && j < inLine)
        stream << *itrLast;
    else
        stream << "...";

    stream << ']'
           << "(total " << lids.size() << ')';

    IBDIAG_RETURN_VOID;
}

bool FLIDsManager::GetFLID(const IBPort &port, lid_t &flid) const
{
    IBDIAGNET_ENTER;

    const SMP_PortInfo *p_portInfo =
                    this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPPortInfo(port.createIndex);
    if (!p_portInfo)
        IBDIAG_RETURN(false);

    //u_int64_t GIDPrfx;  layout: [rest... ][flid: 16 bit][prefix: 16 bit]
    flid = static_cast<lid_t>(p_portInfo->GIDPrfx >> 16) & 0x0ffff;

    IBDIAG_RETURN(true);
}

int FLIDsManager::CheckSwitches(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    for (lid_to_nodes_map::const_iterator itr = this->swicthesByFLID.begin();
            itr != this->swicthesByFLID.end(); itr++) {
        const vec_const_pnode &switches = itr->second;

        if (switches.size() <= 1)
           continue;

        if (switches[0] == NULL || switches[1] == NULL){
            this->error = "DB error - found null node in FLIDs to Switches map";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        stringstream ss;
        ss << "Same FLID= " << itr->first
           << " detected on the switches: "
           << switches[0]->getName() << ", " << switches[1]->getName()
           << (switches.size() > 2 ? "..." : ".")
           <<  " Probably compression ratio is not 1:1.";

        errorsList.push_back(new FLIDError(ss.str()));
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::CollectEnabledFLIDs()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    this->error = "";

    for (set_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->Routers.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->Routers.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->error = "DB error - found null node in Routers set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        const SMP_RouterInfo *p_router_info =
            this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterInfo(p_curr_node->createIndex);
        if (!p_router_info)
            continue;

        if (!this->IsConfiguredFLID(*p_curr_node, *p_router_info))
            continue;

        if (p_router_info->local_router_lid_top || p_router_info->local_router_lid_base) {
            if (p_router_info->global_router_lid_base > p_router_info->local_router_lid_base ||
                    p_router_info->local_router_lid_top > p_router_info->global_router_lid_top) {

                if (this->error.empty())
                    this->error = "Some of the routers have wrongly defined local and global ranges";

                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }

            //collect NON-Local enabled FLIDs
            this->CollectEnabledFLIDs(p_router_info->global_router_lid_base,
                                      p_router_info->local_router_lid_base - 1,
                                      *p_curr_node);
            this->CollectEnabledFLIDs(p_router_info->local_router_lid_top + 1,
                                      p_router_info->global_router_lid_top,
                                      *p_curr_node);
        } else
            //collect NON-Local enabled FLIDs
            this->CollectEnabledFLIDs(p_router_info->global_router_lid_base,
                                      p_router_info->global_router_lid_top,
                                      *p_curr_node);
    }

    IBDIAG_RETURN(rc);
}

void FLIDsManager::CollectEnabledFLIDs(u_int32_t from, u_int32_t to, IBNode &router) const
{
    IBDIAGNET_ENTER;

    const size_t size = ARRAY_SIZE(SMP_RouterLIDTable::V);
    uint8_t block_start = static_cast<uint8_t>(from / size);
    uint8_t block_end = static_cast<uint8_t>(to / size);

    for (uint8_t block = block_start; block <= block_end; block++) {
        const SMP_RouterLIDTable *p_router_lid_table =
                this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterLIDTbl(router.createIndex,
                                                                             block);
        if (!p_router_lid_table)
            continue;

        for (size_t i = 0; i < size; i++) {
            lid_t flid = static_cast<lid_t>(block * size + i);
            if (flid < from)
                continue;

            if (flid > to)
                break;

            if (p_router_lid_table->V[i]) {
                //fill in data for CA to FLID path verification
                //NON-local FLIDs are checked
                this->p_ibdiag->GetDiscoverFabricPtr()->RoutersByFLID[flid].push_back(&router);
                router.enabled_flids.insert(flid);
            }
        }
    }

    IBDIAG_RETURN_VOID;
}

int FLIDsManager::Dump(const string &name)
{
    ofstream  outStream;
    int rc = this->p_ibdiag->OpenFile("FLID", name, outStream, false, true);
    if (rc) {
        this->error = p_ibdiag->GetLastError();
        IBDIAG_RETURN(rc);
    }

    if (!outStream.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    rc = this->Dump(outStream);

    this->p_ibdiag->CloseFile(outStream);

    IBDIAG_RETURN(rc);
}

int FLIDsManager::Dump(ostream &outStream)
{
    IBDIAGNET_ENTER;

    //ranges should not be empty if we are here
    int rc = this->DumpRanges("Global FLID range", this->globalRanges, outStream);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = this->DumpRanges("Local subnet FLID range", this->localRanges, outStream);
    if (rc)
        IBDIAG_RETURN(rc);

    this->DumpCommonLids(outStream);

    outStream << endl
            <<"------------------------------------------------------"
            << endl;

    rc = this->DumpAdjSubnets(outStream);
    if (rc)
        IBDIAG_RETURN(rc);

    outStream << endl
            <<"------------------------------------------------------"
            << endl;

    rc = this->DumpEnabledFLIDsOnRouters(outStream);
    if (rc)
        IBDIAG_RETURN(rc);

    outStream << endl
            <<"------------------------------------------------------"
            << endl;

    rc = this->DumpFLIDsPerSwicthes(outStream);
    if (rc)
        IBDIAG_RETURN(rc);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FLIDsManager::DumpCommonLids(ostream &stream) const
{
     if (!this->common.empty()) {
        stream << "Local LIDs";
        this->LidsToStream(this->common, stream, -1);
        stream << " found in Global FLIDs range"
                  << endl;
    }
    else {
        stream << "Local subnet LID and global FLID ranges are OK"
            << endl;
    }
}

int FLIDsManager::DumpRanges(const string &message, const ranges_map &ranges, ostream &outStream)
{
    IBDIAGNET_ENTER;

    if (ranges.empty()) {
        outStream << message << ": start=" << 0
                << " end=" << 0
                << endl;
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    if (ranges.size() == 1) {
        outStream <<  message << ": start=" << ranges.begin()->first.start
                << " end=" << ranges.begin()->first.end
                << endl;
       IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    outStream << "different " << message << "s found on routers:" << endl;
    int rc = this->RangesToStream(ranges, outStream, -1);
    outStream << endl;

    IBDIAG_RETURN(rc);
}

int FLIDsManager::DumpAdjSubnets(ostream &outStream)
{
    IBDIAGNET_ENTER;

    outStream << "Adjacent subnets" <<  endl;
    for (adj_subnets_ranges_map::const_iterator itr = this->adjSubnetsRanges.begin();
            itr != this->adjSubnetsRanges.end(); itr++)
    {
        outStream << "SubnetPrefixID: " << PTR(itr->first) << ' ';

        if (this->localSubnetPrefix == itr->first)
            outStream << "(local subnet) ";

        int rc = this->DumpRanges("FLID range", itr->second, outStream);
        if (rc)
            IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FLIDsManager::DumpEnabledFLIDsOnRouters(ostream &outStream)
{
    IBDIAGNET_ENTER;

    outStream << endl << "Routers" << endl;

    for (set_pnode::const_iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->Routers.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->Routers.end(); ++nI) {

        const IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->error = "DB error - found null node in Routers set";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        const SMP_RouterInfo *p_router_info =
            this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterInfo(p_curr_node->createIndex);

        if (!p_router_info ||
                !this->IsConfiguredFLID(*p_curr_node, *p_router_info))
            continue;

        outStream << PTR(p_curr_node->guid_get())
                  << " - " << '\"' <<p_curr_node->getName() << '\"'
                  << ", enabled FLIDs"
                  << endl;

        this->LocalEnabledFLIDsToStream(*p_curr_node, *p_router_info, outStream);
        this->NonLocalEnabledFLIDsToStream(*p_curr_node, outStream);

        outStream << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FLIDsManager::LocalEnabledFLIDsToStream(const IBNode &router,
                                             const SMP_RouterInfo &routerInfo, ostream &outStream)
{
    IBDIAGNET_ENTER;

    outStream << "local:"
              << endl;

    const size_t size = ARRAY_SIZE(SMP_RouterLIDTable::V);
    //take care of a case when local flid range is wrongly defined and contains the global one.
    uint8_t block_start = static_cast<uint8_t>(std::max(routerInfo.global_router_lid_base,
                                                        routerInfo.local_router_lid_base) / size);
    uint8_t block_end = static_cast<uint8_t>(std::min(routerInfo.global_router_lid_top,
                                                      routerInfo.local_router_lid_top) / size);
    for (uint8_t block = block_start; block <= block_end; block++) {
        const SMP_RouterLIDTable *p_router_lid_table =
                    this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPRouterLIDTbl(
                                                                        router.createIndex,
                                                                        block);
        if (!p_router_lid_table)
            continue;

        for (size_t i = 0; i < size; i++) {
            lid_t flid = static_cast<lid_t>(block * size + i);
            if (flid < routerInfo.local_router_lid_base || flid < routerInfo.global_router_lid_base)
                continue;

            if (flid > routerInfo.local_router_lid_top || flid > routerInfo.global_router_lid_top)
                break;

            if (p_router_lid_table->V[i])
                outStream << flid
                          << endl;
        }

        outStream << endl;
    }

    IBDIAG_RETURN_VOID;
}

void FLIDsManager::NonLocalEnabledFLIDsToStream(const IBNode &router, ostream &outStream)
{
    IBDIAGNET_ENTER;

    outStream << "non-local:"
              << endl;

    for (set_lid::const_iterator itr = router.enabled_flids.begin();
            itr != router.enabled_flids.end(); itr++) {
        outStream << *itr
                  << endl;
    }

    IBDIAG_RETURN_VOID;
}

int FLIDsManager::DumpFLIDsPerSwicthes(ostream &outStream)
{
    IBDIAGNET_ENTER;

    outStream << endl << "FLID per switches" << endl;
    for (lid_to_nodes_map::const_iterator itr = this->swicthesByFLID.begin();
            itr != this->swicthesByFLID.end(); itr++){

        for (vec_const_pnode::const_iterator itrSwitch = itr->second.begin();
                itrSwitch != itr->second.end(); itrSwitch++) {

            const IBNode *p_swicth = *itrSwitch;
            if (!p_swicth) {
                this->error = "DB error - Null pointer found in switches list";
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            outStream << PTR(p_swicth->guid_get())         << " - "
                    << '\"' << p_swicth->getName() << '\"' << ": "
                    << itr->first
                    << endl;
        }
    }

    if (this->swicthesWithMultipleFLIDs.empty())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    outStream << endl << "Different FLIDs were detected on the following switches" << endl;
    for (node_to_lids_map::const_iterator itr = this->swicthesWithMultipleFLIDs.begin();
            itr != this->swicthesWithMultipleFLIDs.end(); itr++) {

        const IBNode *p_swicth = itr->first;
        if (!p_swicth) {
            this->error = "DB error - Null pointer found in node to flids map";
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        outStream << PTR(p_swicth->guid_get()) << " - "
                << '\"' << p_swicth->getName() << '\"' << ':' << endl;
        int rc = this->FLIDsToStream(itr->second, outStream, -1);
        if (rc)
            IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
