/*
 *  Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 *  Copyright (c) 2022-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#ifndef _IBDIAGAPP_H_
#define _IBDIAGAPP_H_

#include <stdlib.h>

#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>
#include <misc/cmdparser/cmdparser.h>

#include "ibdiagapp_types.h"
#include "ibdiagnet_stages.h"
#include "ibdiagnet_plugins.h"
#include "ibdiagapp_lc_plugin.h"
#include "ibdiagnet_debug_wrap.h"

/******************************************************/
class IBDiagapp : public CommandLineRequester {
private:
    std::stringstream m_stage_info_csv;

protected:
    ////////////////////
    //members
    ////////////////////
    typedef struct skip_obj {
        u_int32_t value;
        string name;
    } skip_obj_t;

    vector <struct skip_obj> skip_options_list;

    enum {FAILED_TO_CONSTRUCT, NEED_TO_GET_ARGS, NOT_INITILIAZED, READY} ibdiagnet_status;
    class Stage *stages[NUM_OF_STAGES];
    list_p_symbols_plugin loaded_plugins_list;
    list_string not_to_load_plugins_list;
    IBDiag ibdiag_obj;
    CommandLineParser *p_main_parser;

    unsigned int num_of_errors_to_screen;
    u_int32_t m_max_smps_on_wire;
    u_int32_t m_max_gmps_on_wire;

    u_int64_t max_iterations;
    u_int64_t timeout_iteration_mic_sec;

    string base_path;

    CSVOut csv_out;

    LCPluginMgr m_lc_plugins_mgr;   //the manager of loosely coupled plugins

    FilesBasedMKeyManager *p_mkeymngr;

    bool is_default_config_file;

    string config_file;
    string generated_config_file;
    string mkey_val;
    string mkey_path;
    string am_key_path;
    string vs_key_path;
    string cc_key_path;
    string m2n_key_path;

    DebugWrap m_debug_wraper;

    ////////////////////
    //methods
    ////////////////////
    int ParseSkipList(const char *skip_list);
    int ParseEnableDisableOutput(const string & lst, bool value);
    int ParsePathPerFileOutput(const string & lst);
    int ParsePOption(const char *option_p_list, map_str_uint64& counters_to_threshold_map);
    int ParseSkipPluginList(const char *skip_plugin_list);
    int ParseKeyValue(const string key_str, u_int64_t &key);
    int ParseMKeyValue();
    int ParseMKeyFiles();
    int ParseConfigFile();
    int ParseGuid2Key(const string &guid_2_key_file_path, map_guid_2_key_t &guid_2_key,
                      string &key_type_str, bool should_print_error,
                      const string &default_file_name = "");
    int CreateConfigFile(string file_path);
    int HandleMKeyParams();
    void PrintFabricSummaryTable();
    void PrintSummary();

    void LoadPlugins();
    void UnloadPlugins();

    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit
    int GetLocalGUIDFromUser(u_int64_t& guid);

    //Returns: 0 - OK / 1 - error
    int DumpStartInfoCSVTable(int argc, const char **argv);

    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit / 3 - parsing error
    virtual int HandleOption(string name, string value);
    virtual int PostParseFiles();

    virtual Stage* CreateStage(uint8_t stageIndex) = 0;
    virtual string GetToolName() = 0;

    void OpenPcapFile();

public:
    ////////////////////
    //methods
    ////////////////////
    IBDiagapp();
    virtual ~IBDiagapp();

    int PostInit();
    void SetIBNLDir(const string &path);
    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit
    int GetOpt(int argc, const char **argv);

    //Returns: 0 - OK / 1 - error
    virtual int Init();
    int RunAllStages();
    int RunStages(int from, int to);
    int RunSingleStage(Stage* stage);
    int Run(int argc, const char **argv);

    int IsValidLid(lid_t lid) {
        if(lid > IB_MAX_UCAST_LID) {
            PRINT("lid should be less than %d\n",
                  IB_MAX_UCAST_LID);
            IBDIAGNET_RETURN(1);
        }
        IBDIAGNET_RETURN(0);
    }

    bool isValidLoadFromFilePath();

    inline string get_version()
    {
        return string("\"" IBDIAGNET_VERSION "\",") + get_ibdiag_version();
    }

    inline void SetAMKeyPath(string am_key_path) {
        this->am_key_path = am_key_path;
    }
    inline const string& GetAMKeyPath() const { return this->am_key_path; }

    inline void SetVSKeyPath(string vs_key_path, bool forced) {
        if (forced || this->vs_key_path.empty())
            this->vs_key_path = vs_key_path;
    }
    inline const string& GetVSKeyPath() const { return vs_key_path; }

    inline void SetCCKeyPath(string cc_key_path, bool forced) {
        if (forced || this->cc_key_path.empty())
            this->cc_key_path = cc_key_path;
    }
    inline const string& GetCCKeyPath() const { return cc_key_path; }

    // Class 0xC M2N Key
    inline void SetM2NKeyPath(string m2n_key_path, bool forced) {
        if (forced || this->m2n_key_path.empty())
            this->m2n_key_path = m2n_key_path;
    }
    inline const string& GetM2NKeyPath() const { return m2n_key_path; }

    inline void SetMaxIterations(u_int64_t max_iterations) {
        this->max_iterations = max_iterations;
    }
    inline u_int64_t GetMaxIterations() const { return this->max_iterations; }
    inline void SetTimeoutIteration(u_int64_t timeout_iteration_mil_sec) {
        //convert to mic seconds
        this->timeout_iteration_mic_sec = timeout_iteration_mil_sec * 1000;
    }
    inline u_int64_t GetTimeoutIteration() const { return this->timeout_iteration_mic_sec; }
    void TimeoutActivation(struct timeval *iteration_time);
    inline void SetConfigFilePath(const string &config_file) {
        this->config_file = config_file;
        this->is_default_config_file = false;
    }
};


#endif          /* not defined _IBDIAGAPP_H_ */
