/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "ibdiagnet_types.h"
#include <string.h>
#include <unistd.h>
extern char *optarg;
extern int optind, opterr, optopt;
#include <getopt.h>
#include <errno.h>

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>

#include "ibdiagnet.h"
#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>

/****************************************************/
/* globals */
/****************************************************/


/****************************************************/
/*******************public methods*******************/
/****************************************************/

IBDiagnet::IBDiagnet()
{
}

Stage* IBDiagnet::CreateStage(uint8_t stageIndex) {

    IBDIAGNET_ENTER;
    Stage *stage = NULL;
    switch (stageIndex) {
        case STAGE_IDX_DISCOVERY:
            //discovery stage
            stage = new DiscoveryFabricStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_DISCOVERY);
            break;
        case STAGE_IDX_LIDS:
            //lids stage
            stage = new LidsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LIDS);
            break;
        case STAGE_IDX_TOPOLOGY_MATCH:
            //topology match stage
            stage = new TopologyMatchStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_TOPOLOGY_MATCH);
            break;
        case STAGE_IDX_LINKS:
            //links stage
            stage = new LinksStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LINKS);
            break;
        case STAGE_IDX_SM:
            //sm stage
            stage = new SMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SM);
            break;
        case STAGE_IDX_PM:
            //pm stage
            stage = new PMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PM);
            break;
       case STAGE_IDX_ROUTING:
            //routing stage
            stage = new RoutingStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_ROUTING);
            break;
        case STAGE_IDX_NODES_INFO:
            //nodes info stage
            stage = new NodesInfoStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_NODES_INFO);
            break;
        case STAGE_IDX_SPEED_WIDTH:
            //speed / width stage
            stage = new Speed_WidthStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SPEED_WIDTH);
            break;
        case STAGE_IDX_PKEY:
            //pkey stage
            stage = new PKeyStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PKEY);
            break;
        case STAGE_IDX_AGUID:
            //aguid stage
            stage = new AGUIDStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_AGUID);
            break;
        case STAGE_IDX_MLNX_CNTRS:
            //Mellanox Counters stage
            stage = new MlnxCntrsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_MLNX_CNTRS);
            break;
        case STAGE_IDX_VIRTUALIZATION:
            //Virtualiztion stage
            stage = new VirtualizationStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_VIRTUALIZATION);
            break;
        case STAGE_IDX_TEMPSENSING:
            //Temperature stage
            stage = new TemperatureSensing(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_TEMPSENSING);
            break;
        case STAGE_IDX_ROUTERS:
            stage = new RoutersStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_ROUTERS);
            break;
        case STAGE_IDX_SHARP:
           //Sharp stage
           stage = new SharpStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                         STAGE_NAME_SHARP);
           break;
        case STAGE_IDX_QOS:
           //QoS stage
           stage = new QoSStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                         STAGE_NAME_QOS);
           break;
        case STAGE_IDX_CC:
            // CC stage
            stage = new CongestionControlStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_CC);
           break;

        case STAGE_IDX_FAST_RECOVERY:
            // fast recovery stage
            stage = new FastRecoveryStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_FAST_RECOVERY);
            break;

        case STAGE_IDX_RAIL_OPT_TOPO_CHECK:
            // CC stage
            stage = new CheckRailOptimizedTopologyStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_RAIL_OPT_TOPO_CHECK);
           break;
        case STAGE_IDX_POST_REPORTS:
            stage = new PostReportsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_POST_REPORTS);
           break;
        case STAGE_IDX_DFP_TOPO_CHECK:
            stage = new CheckDFPTopologyStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_DFP_TOPO_CHECK);
           break;
        case STAGE_IDX_FAT_TREE_TOPO_CHECK:
            stage = new CheckFatTreeTopologyStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_FAT_TREE_TOPO_CHECK);
           break;
        case STAGE_IDX_REPORT_CABLE_INFO:
            stage = new ReportCableInfoStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_REPORT_CABLE_INFO);
            break;
        case STAGE_IDX_EXPORT_DATA:
            stage = new ExportDataStage(&this->ibdiag_obj, &this->loaded_plugins_list);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_EXPORT_DATA);
        default:
            break;
    }

    IBDIAGNET_RETURN(stage);
}

IBDiagnet::~IBDiagnet()
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN_VOID;
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagnet::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    int res = IBDiagapp::HandleOption(name, value);

    /* if res is 1 so it's unknown stage to ibdiagapp, need to
     * check for ibdiagnet stages */

    bool bool_flag_val = true;

    if (res == 1) {
        if (name == OPTION_CREATE_CONFIG) {
            this->LoadPlugins();
            this->generated_config_file = value;
            if (this->CreateConfigFile(this->generated_config_file)) {
                IBDIAGNET_RETURN(3);
            }
            IBDIAGNET_RETURN(2);           //exit with no error
        } else if (name == OPTION_VLR) {
            ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetPSLDumpFile(value);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->UseVLR();
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SA_DUMP) {
            ((AGUIDStage *)this->stages[STAGE_IDX_AGUID])->MarkStageAsActive();
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                SetSADumpFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SMDB) {
            this->ibdiag_obj.SetSMDBPath(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_ROUTING) {
            this->ParseBoolValue(value, bool_flag_val);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                    MarkStage(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_ROUTING_OPT) {
            if (!this->stages[STAGE_IDX_ROUTING]->IsActive()){
                ERR_PRINT_ARGS("Use %s option only after %s option.\n",
                        OPTION_ROUTING_OPT, OPTION_ROUTING);
                IBDIAGNET_RETURN(3);
            }
            string opt_str = "," + value + ",";

            if (opt_str.find(",skip_vs,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->SkipVSRoute();
                INFO_PRINT("Skipping \"vs\" (vendor specific) routing option.\n");
            }
            if (opt_str.find(",mcast,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                    CheckMcastCredLoop();
                INFO_PRINT("Using \"mcast\" (mcast credit loop check) routing \
                            option.\n");
            }
            if (opt_str.find(",check_sl,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                    SetCheckSL();
                INFO_PRINT("Using \"check_sl\" routing option.\n");
            }
            if (opt_str.find(",sl=") != std::string::npos) {
                size_t offset = opt_str.find(",sl=") + strlen(",sl=");
                string value_str = opt_str.substr(offset);
                 //take off ',' from value_str
                size_t pos = value_str.find_first_of(',');
                if (pos != std::string::npos)
                   value_str = value_str.substr(0, pos);

                u_int8_t sl = (u_int8_t)strtoul(value_str.c_str(), NULL, 0);
                if (!is_non_neg_num(value_str) ||  sl >= IB_NUM_SL) {
                    ERR_PRINT_ARGS("Illegal value for \'--%s\' option: sl=%s\n",
                            OPTION_ROUTING_OPT, value_str.c_str());
                } else {
                    ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                        SetDefaultSL(sl);
                    INFO_PRINT("Using \"sl=%d\" routing option.\n", sl);
                }
            }
            if (opt_str.find(",dump_only,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpOnly();
                INFO_PRINT("Using \"dump_only\" routing option.\n");
            }
            if (opt_str.find(",dump_only_skip_routing_tables,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpOnlySkipRoutingTables();
                INFO_PRINT("Using \"dump_only_skip_routing_tables\" routing option.\n");
            }
            if (opt_str.find(",static_ca2ca,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->StaticCA2CA();
                INFO_PRINT("Using \"static_ca2ca\" routing option.\n");
            }
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_FATTREE) {
            //this option is going to be deleted in the future
            this->ParseBoolValue(value, bool_flag_val);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->SetFatTree(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_WRITE) {
            ((TopologyMatchStage *)this->stages[STAGE_IDX_TOPOLOGY_MATCH])->
                SetOutputTopoFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_MATCH_FILE) {
            ((TopologyMatchStage *)this->stages[STAGE_IDX_TOPOLOGY_MATCH])->
                SetInputTopoFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_IBNL_DIR) {
        	SetIBNLDir(value);
            ((TopologyMatchStage *)this->stages[STAGE_IDX_TOPOLOGY_MATCH])->SetTopoIbnlDir(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SCOPE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetScopePortGuidFile(value);

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_EXCLUDE_SCOPE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetExcludeScopePortGuidFile(value);

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_HEALTHY_PORTS_POLICY) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetHealthyPortsPolicyFile(value);

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_UNHEALTHY_PORTS_DUMP) {
            DiscoveryFabricStage *p_stage =
                        (DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY];
            if (p_stage->GetHealthyPortsPolicyFile().empty()) {
                ERR_PRINT_ARGS("Use the %s option only after %s option.\n",
                        OPTION_UNHEALTHY_PORTS_OPT, OPTION_HEALTHY_PORTS_POLICY);
                IBDIAGNET_RETURN(3);
            }

            p_stage->SetUnhealthyPortsDumpFile(value);

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_UNHEALTHY_PORTS_OPT) {
            DiscoveryFabricStage *p_stage =
                        (DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY];
            if (p_stage->GetHealthyPortsPolicyFile().empty() ||
                    p_stage->GetUnhealthyPortsDumpFile().empty()) {
                ERR_PRINT_ARGS("Use the %s option only after %s and %s options.\n",
                        OPTION_UNHEALTHY_PORTS_OPT,
                        OPTION_HEALTHY_PORTS_POLICY, OPTION_UNHEALTHY_PORTS_DUMP);
                IBDIAGNET_RETURN(3);
            }

            string switch_action;
            string ca_action;
            transform(value.begin(), value.end(), value.begin(), ::tolower);
            ParseUnhealthyPortsOption(value, switch_action, ca_action);

            if (!p_stage->SetUnhealthyPortsSwitchAction(switch_action)) {
                ERR_PRINT_ARGS("Wrong switch action=%s provided for the option %s.\n",
                            switch_action.c_str(), OPTION_UNHEALTHY_PORTS_OPT);
                IBDIAGNET_RETURN(3);
            }

            if (!p_stage->SetUnhealthyPortsCAAction(ca_action)) {
                ERR_PRINT_ARGS("Wrong ca action=%s provided for the option %s.\n",
                            ca_action.c_str(), OPTION_UNHEALTHY_PORTS_OPT);
                IBDIAGNET_RETURN(3);
            }

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHARP) {
            this->ParseBoolValue(value, bool_flag_val);
            ((SharpStage *)this->stages[STAGE_IDX_SHARP])->MarkStage(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHARP_VERSION) {
            u_int8_t version = (u_int8_t)strtoul(value.c_str(), NULL, 0);

            if (!is_non_neg_num(value) || version > SHARP_CONTROL_VERSION_MAX) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                        "    (Legal values: %s).\n",
                        OPTION_SHARP_VERSION, value.c_str(), OPTION_SHARP_VERSION_ARG_DESC);
                IBDIAGNET_RETURN(3);
            }

            ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetVersion(version);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHARP_OPT) {
            if (!this->stages[STAGE_IDX_SHARP]->IsActive()){
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->MarkStageAsActive();
            }
            string opt_str = "," + value + ",";
            if (opt_str.find(",csc,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetClearCounters();
                INFO_PRINT("Using \"csc\" (clear sharp counters) sharp option.\n");
            }
            if (opt_str.find(",dsc,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetDumpCounters();
                INFO_PRINT("Using \"dsc\" (dump sharp counters) sharp option.\n");
            }
            if (opt_str.find(",dscp,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetDumpHBAPortCounters();
                INFO_PRINT("Using \"dscp\" (dump sharp HBA PORT counters) sharp option.\n");
            }
            if (opt_str.find(",ad_hoc,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetAdHoc();
                INFO_PRINT("Using \"ad_hoc\" sharp option.\n");
            }
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_AM_KEY) {
            u_int64_t am_key;
            if (!this->ParseKeyValue(value, am_key))
                IBKey<IB_AM_KEY>::SetDefaultKey(am_key);
            else
                ERR_PRINT("Failed to parse AM Key: %s\n", value.c_str());
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_AM_KEY_FILE) {
            this->SetAMKeyPath(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_LOAD_FROM_FILE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->SetLoadFromFilePath(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_RAIL_TOPO_CHECK) {
            this->ParseBoolValue(value, bool_flag_val);
            if (bool_flag_val) {
                CheckRailOptimizedTopologyStage *p_stage =
                        (CheckRailOptimizedTopologyStage *)this->stages[STAGE_IDX_RAIL_OPT_TOPO_CHECK];
                p_stage->MarkStageAsActive();
                p_stage->SetRailFlow(1); //we keep it as integer for future development
                this->ibdiag_obj.SetRailValidation(true);
            }

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_RAIL_TOPO_CHECK_OPTIONS) {
            if (!this->stages[STAGE_IDX_RAIL_OPT_TOPO_CHECK]->IsActive()) {
                ERR_PRINT_ARGS("Use %s option only after %s option.\n",
                        OPTION_RAIL_TOPO_CHECK_OPTIONS, OPTION_RAIL_TOPO_CHECK);
                IBDIAGNET_RETURN(3);
            }

            if (!this->ParseRailOptions(value)) {
                IBDIAGNET_RETURN(3);
            }

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_RAIL_BDF_SOURCE) {
            if (!is_non_neg_num(value)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\': %s. Possible values are: 0, 1, 2\n",
                               OPTION_RAIL_BDF_SOURCE, value.c_str());
                IBDIAGNET_RETURN(3);
            }

            size_t mode = strtoul(value.c_str(), NULL, 0);
            if (mode > 2) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\': %s. Possible values are: 0, 1, 2\n",
                               OPTION_RAIL_BDF_SOURCE, value.c_str());
                IBDIAGNET_RETURN(3);
            }

            ((CheckRailOptimizedTopologyStage *)this->stages[STAGE_IDX_RAIL_OPT_TOPO_CHECK])->
                    SetBDFMode(static_cast<PCI_BDF_SOURCE>(mode));

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_RAIL_BDF_FACTOR) {
            if (!is_non_neg_num(value)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\': %s. Possible values are from 0 to 100\n",
                               OPTION_RAIL_BDF_FACTOR, value.c_str());
                IBDIAGNET_RETURN(3);
            }

            size_t factor = strtoul(value.c_str(), NULL, 0);
            if (factor > 100) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\': %s. Possible values are from 0 to 100\n",
                               OPTION_RAIL_BDF_FACTOR, value.c_str());
                IBDIAGNET_RETURN(3);
            }

            ((CheckRailOptimizedTopologyStage *)this->stages[STAGE_IDX_RAIL_OPT_TOPO_CHECK])->
                    SetBDFFactor(factor);

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_DFP) {
            this->ParseBoolValue(value, bool_flag_val);
            ((CheckDFPTopologyStage *)this->stages[STAGE_IDX_DFP_TOPO_CHECK])->MarkStage(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_DFP_OPT) {
            if (!this->stages[STAGE_IDX_DFP_TOPO_CHECK]->IsActive()){
                ERR_PRINT_ARGS("Use %s option only after %s option.\n",
                        OPTION_DFP_OPT, OPTION_DFP);
                IBDIAGNET_RETURN(3);
            }
            string opt_str = "," + value + ",";
            if (opt_str.find(",max_cas=") != std::string::npos) {
                size_t offset = opt_str.find(",max_cas=") + strlen(",max_cas=");
                string value_str = opt_str.substr(offset);
                //take off ',' from value_str
                size_t pos = value_str.find_first_of(',');
                if (pos != std::string::npos)
                   value_str = value_str.substr(0, pos);

                if (!is_non_neg_num(value_str)) {
                    ERR_PRINT_ARGS("Illegal value for \'--%s\' option: max_cas=%s\n",
                                OPTION_DFP_OPT, value_str.c_str());
                    IBDIAGNET_RETURN(3);
                }

                ((CheckDFPTopologyStage *)this->stages[STAGE_IDX_DFP_TOPO_CHECK])->
                                    SetMaxCAsOnSpine((int)strtoul(value_str.c_str(), NULL, 0));
            }
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_FAT_TREE) {
                this->ParseBoolValue(value, bool_flag_val);
                ((CheckFatTreeTopologyStage *)this->stages[STAGE_IDX_FAT_TREE_TOPO_CHECK])->
                                    MarkStage(bool_flag_val);
                IBDIAGNET_RETURN(0);
        } else if (name == OPTION_FAT_TREE_ROOTS_REGEX) {
            if (!this->stages[STAGE_IDX_FAT_TREE_TOPO_CHECK]->IsActive()) {
                ERR_PRINT_ARGS("Use %s option only after %s option.\n",
                        OPTION_FAT_TREE_ROOTS_REGEX, OPTION_FAT_TREE);
                IBDIAGNET_RETURN(3);
            }

            if (!((CheckFatTreeTopologyStage*)
                        (this->stages[STAGE_IDX_FAT_TREE_TOPO_CHECK]))->SetRootsRegex(value)) {
                ERR_PRINT_ARGS("Invalid regular expression: '%s'" 
                            " for Fat-Tree roots was provided.\n", value.c_str());
                IBDIAGNET_RETURN(3);
            }

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_AGUID) {
            if (!((AGUIDStage *)this->stages[STAGE_IDX_AGUID])->IsActive()) {
                this->ParseBoolValue(value, bool_flag_val);
                ((AGUIDStage *)this->stages[STAGE_IDX_AGUID])->MarkStage(bool_flag_val);
            }
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_SMART_NAMES) {
            this->ParseBoolValue(value, bool_flag_val);
            ((TopologyMatchStage *)this->stages[STAGE_IDX_TOPOLOGY_MATCH])->
                        SetSmartNames(bool_flag_val);

            IBDIAGNET_RETURN(0);
        }
        else if (name == OPTION_EXPORT) {
            if (this->ibdiag_obj.InitExportAPI(value))
                IBDIAGNET_RETURN(3);
            this->stages[STAGE_IDX_EXPORT_DATA]->MarkStageAsActive();
            IBDIAGNET_RETURN(0);
        }
        else if (name == OPTION_CONTROL) {
            if (this->ibdiag_obj.InitControlAPI(value))
                IBDIAGNET_RETURN(3);
            IBDIAGNET_RETURN(0);
        }
    }

    IBDIAGNET_RETURN(res);
}

int IBDiagnet::PostParseFiles()
{
    IBDIAGNET_ENTER;

    int rc = IBDiagapp::PostParseFiles();
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (!this->GetAMKeyPath().empty()) {
        string key_type_str = "AM Key";
        map_guid_2_key_t &am_key_map =
                     ((SharpStage *)this->stages[STAGE_IDX_SHARP])->GetAMKeyMap();
        rc = this->ParseGuid2Key(this->GetAMKeyPath(), am_key_map, key_type_str, true);
        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED) {
            IBDIAGNET_RETURN(3);
        } else if (rc) {
            WARN_PRINT("Parse %s file was finished with warnings\n\n", key_type_str.c_str());
        } else {
            INFO_PRINT("Parse %s file was finished successfully\n\n", key_type_str.c_str());
        }
    }

    CheckDFPTopologyStage *dfp_stage =
        (CheckDFPTopologyStage *)this->stages[STAGE_IDX_DFP_TOPO_CHECK];

    if (!this->ibdiag_obj.GetSMDBPath().empty()) {
        if (this->ibdiag_obj.ParseSMDBFile()) {
            ERR_PRINT("Parsing SMDB file: %s was finished with errors\n",
                      this->ibdiag_obj.GetSMDBPath().c_str());
            IBDIAGNET_RETURN(3);
        }

        INFO_PRINT("Parsing SMDB file: %s was finished successfully\n",
                   this->ibdiag_obj.GetSMDBPath().c_str());

        const IBRoutingEngine routing_engine = this->ibdiag_obj.GetIBDiagSMDB().GetRoutingEngine();

        if (dfp_stage->IsActive()) {
            if (routing_engine != IB_ROUTING_DFP) {
                dfp_stage->MarkStage(false);
                WARN_PRINT(STAGE_NAME_DFP_TOPO_CHECK " stage is skipped, Routing Engine is %s\n",
                           routing_engine2char(routing_engine));
            } else {
                dfp_stage->SetIsRootFromFile(true);
                dfp_stage->SetMaxCAsOnSpine(-1);
            }
        }

        CheckFatTreeTopologyStage *ft_stage =
                    (CheckFatTreeTopologyStage*) this->stages[STAGE_IDX_FAT_TREE_TOPO_CHECK];
        if (ft_stage->IsActive()) {
            //todo what is proper engine type?
            if (routing_engine == IB_ROUTING_FTREE    ||
                routing_engine == IB_ROUTING_AR_FTREE ||
                routing_engine == IB_ROUTING_UPDN     ||
                routing_engine == IB_ROUTING_AR_UPDN)
                ft_stage->SetRootsFromFile(true);
            else
                WARN_PRINT("The Routing Engine %s found in the smdb file should "
                           "not be applied for Fat-Tree topologies. "
                           "The file will not be used for roots detection.\n",
                           routing_engine2char(routing_engine));
        }
    } else {
    	if (dfp_stage->IsActive() && dfp_stage->GetMaxCAsOnSpine() == -1) {
                dfp_stage->SetMaxCAsOnSpine(1);
    	}
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

static string get_pattern(const string &value, const string &pattern) {
    string data;
    size_t delim = 0;
    size_t found = value.find(pattern);
    if (found != std::string::npos) {
        delim = value.find(',', found);
        if (delim != std::string::npos)
            data = value.substr(found + pattern.size(), delim - pattern.size());
        else
            data = value.substr(found + pattern.size());
    }

    return data;
}

void IBDiagnet::ParseUnhealthyPortsOption(const string &value,
                                          string &switch_action, string &ca_action)
{
    IBDIAGNET_ENTER;

    // the input variable is already converted to lower case.
    switch_action = get_pattern(value, "switch=");
    ca_action = get_pattern(value, "ca=");

    IBDIAGNET_RETURN_VOID;
}

bool IBDiagnet::ParseRailOptions(const string& value)
{
    IBDIAGNET_ENTER;

    const string pattern = RAIL_OPTION_REG_EXP;
    size_t found = value.find(pattern);
    if (found == std::string::npos ||
        found + pattern.size() == value.size()) {
        ERR_PRINT_ARGS("regex is not specified correctly.\n");
        IBDIAGNET_RETURN(false);
    }

    string reg_exp = value.substr(found + pattern.size());
    if (!((CheckRailOptimizedTopologyStage *)this->stages[STAGE_IDX_RAIL_OPT_TOPO_CHECK])->
                    SetRailRegex(reg_exp)) {
       ERR_PRINT_ARGS("The specified regex value: '%s' is not valid.\n",
                reg_exp.c_str());
       IBDIAGNET_RETURN(false);
    }

    IBDIAGNET_RETURN(true);
}

int IBDiagnet::Init()
{
    IBDIAGNET_ENTER;
    int res;

    skip_obj_t obj0 = {SKIP_DUP_GUIDS_VALUE, SKIP_DUP_GUIDS_DESC_PRINT};
    skip_options_list.push_back(obj0);
    skip_obj_t obj1 = {SKIP_DUP_NODE_DESC_VALUE, "dup_node_desc"};
    skip_options_list.push_back(obj1);
    skip_obj_t obj2 = {SKIP_STAGE_LIDS_VALUE, "lids"};
    skip_options_list.push_back(obj2);
    skip_obj_t obj3 = {SKIP_STAGE_SM_VALUE, "sm"};
    skip_options_list.push_back(obj3);
    skip_obj_t obj4 = {SKIP_STAGE_NODES_INFO_VALUE, "nodes_info"};
    skip_options_list.push_back(obj4);
    skip_obj_t obj5 = {SKIP_STAGE_PKEY_CHECK, "pkey"};
    skip_options_list.push_back(obj5);

    res = IBDiagapp::Init();
    if (res)
        IBDIAGNET_RETURN(res);

    this->AddDescription(IBDIAGNET_TOOL_DESC);
    this->AddOptions(OPTION_VLR, OPTION_VLR_SHORT, OPTION_VLR_ARG_DESC, OPTION_VLR_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_ROUTING, OPTION_ROUTING_SHORT, OPTION_ROUTING_ARG_DESC,
                     OPTION_ROUTING_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_ROUTING_OPT, OPTION_ROUTING_OPT_SHORT, OPTION_ROUTING_OPT_ARG_DESC,
                     OPTION_ROUTING_OPT_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_SA_DUMP, OPTION_SA_DUMP_SHORT, OPTION_SA_DUMP_ARG_DESC,
                     OPTION_SA_DUMP_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_SMDB, OPTION_SMDB_SHORT, OPTION_SMDB_ARG_DESC, OPTION_SMDB_DESC,
                         OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_FATTREE, OPTION_FATTREE_SHORT, OPTION_FATTREE_ARG_DESC,
                     OPTION_FATTREE_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_SCOPE, OPTION_SCOPE_SHORT, OPTION_SCOPE_ARG_DESC,
                     OPTION_SCOPE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_EXCLUDE_SCOPE, OPTION_EXCLUDE_SCOPE_SHORT,
                     OPTION_EXCLUDE_SCOPE_ARG_DESC, OPTION_EXCLUDE_SCOPE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_HEALTHY_PORTS_POLICY, OPTION_HEALTHY_PORTS_POLICY_SHORT,
                     OPTION_HEALTHY_PORTS_POLICY_ARG_DESC, OPTION_HEALTHY_PORTS_POLICY_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_UNHEALTHY_PORTS_DUMP, OPTION_UNHEALTHY_PORTS_DUMP_SHORT,
                     OPTION_UNHEALTHY_PORTS_DUMP_ARG_DESC, OPTION_UNHEALTHY_PORTS_DUMP_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_UNHEALTHY_PORTS_OPT, OPTION_UNHEALTHY_PORTS_OPT_SHORT,
                     OPTION_UNHEALTHY_PORTS_OPT_ARG_DESC, OPTION_UNHEALTHY_PORTS_OPT_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_TOPO_WRITE, OPTION_TOPO_WRITE_SHORT, OPTION_TOPO_WRITE_ARG_DESC,
                     OPTION_TOPO_WRITE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_TOPO_MATCH_FILE, OPTION_TOPO_MATCH_FILE_SHORT,
                     OPTION_TOPO_MATCH_FILE_ARG_DESC, OPTION_TOPO_MATCH_FILE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_TOPO_IBNL_DIR, OPTION_TOPO_IBNL_DIR_SHORT,
                     OPTION_TOPO_IBNL_DIR_ARG_DESC, OPTION_TOPO_IBNL_DIR_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_SHARP, OPTION_SHARP_SHORT, OPTION_SHARP_ARG_DESC, OPTION_SHARP_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SHARP_VERSION, OPTION_SHARP_VERSION_SHORT, OPTION_SHARP_VERSION_ARG_DESC,
                     OPTION_SHARP_VERSION_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_SHARP_OPT, OPTION_SHARP_OPT_SHORT, OPTION_SHARP_OPT_ARG_DESC,
                     OPTION_SHARP_OPT_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_AM_KEY, OPTION_AM_KEY_SHORT, OPTION_AM_KEY_ARG_DESC,
                     OPTION_AM_KEY_DESC, DEFAULT_KEY_STR);
    this->AddOptions(OPTION_AM_KEY_FILE, OPTION_AM_KEY_FILE_SHORT, OPTION_AM_KEY_FILE_ARG_DESC,
                     OPTION_AM_KEY_FILE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_LOAD_FROM_FILE, OPTION_LOAD_FROM_FILE_SHORT,
                     OPTION_LOAD_FROM_FILE_ARG_DESC, OPTION_LOAD_FROM_FILE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_CREATE_CONFIG, OPTION_CREATE_CONFIG_SHORT,
                     OPTION_CREATE_CONFIG_ARG_DESC, OPTION_CREATE_CONFIG_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_NO_CONF_FILE);
    //rails validation
    this->AddOptions(OPTION_RAIL_TOPO_CHECK, OPTION_RAIL_TOPO_CHECK_SHORT,
                     OPTION_RAIL_TOPO_CHECK_ARG_DESC, OPTION_RAIL_TOPO_CHECK_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_RAIL_TOPO_CHECK_OPTIONS, OPTION_RAIL_TOPO_CHECK_OPTIONS_SHORT,
                     OPTION_RAIL_TOPO_CHECK_OPTIONS_ARG_DESC, OPTION_RAIL_TOPO_CHECK_OPTIONS_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_RAIL_BDF_SOURCE, OPTION_RAIL_BDF_SOURCE_SHORT,
                     OPTION_RAIL_BDF_SOURCE_ARG_DESC, OPTION_RAIL_BDF_SOURCE_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_RAIL_BDF_FACTOR, OPTION_RAIL_BDF_FACTOR_SHORT,
                     OPTION_RAIL_BDF_FACTOR_ARG_DESC, OPTION_RAIL_BDF_FACTOR_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);


    this->AddOptions(OPTION_AGUID, OPTION_AGUID_SHORT, OPTION_AGUID_ARG_DESC, OPTION_AGUID_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    //dfp validation
     this->AddOptions(OPTION_DFP, OPTION_DFP_SHORT, OPTION_DFP_ARG_DESC,
                      OPTION_DFP_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

     this->AddOptions(OPTION_DFP_OPT, OPTION_DFP_OPT_SHORT, OPTION_DFP_OPT_ARG_DESC,
                      OPTION_DFP_OPT_DESC, OPTION_DEF_VAL_NULL);

    //fat-tree validation
    this->AddOptions(OPTION_FAT_TREE, OPTION_FAT_TREE_SHORT, OPTION_FAT_TREE_ARG_DESC,
                     OPTION_FAT_TREE_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_FAT_TREE_ROOTS_REGEX, OPTION_FAT_TREE_ROOTS_REGEX_SHORT,
                    OPTION_FAT_TREE_ROOTS_REGEX_ARG_DESC, OPTION_FAT_TREE_ROOTS_REGEX_DESC,
                    OPTION_DEF_VAL_NULL);

    this->AddOptions(OPTION_TOPO_SMART_NAMES, OPTION_TOPO_SMART_NAMES_SHORT,
                     OPTION_TOPO_SMART_NAMES_ARGS_DESC, OPTION_TOPO_SMART_NAMES_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_EXPORT, OPTION_EXPORT_SHORT, OPTION_EXPORT_ARG_DESC, OPTION_EXPORT_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_CONTROL, OPTION_CONTROL_SHORT, OPTION_CONTROL_ARG_DESC,
                     OPTION_CONTROL_DESC, OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);


    IBDIAGNET_RETURN(0);
}


#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
int main (int argc, const char **argv)
{
    OutputControl::app_settings.init("ibdiagnet2");
    OutputControl::instance().enabled().set(AR_FILE, false);
    OutputControl::instance().enabled().set(IBIS_STAT_FILE, false);
    OutputControl::instance().in_summary().set(IBIS_STAT_FILE, false);
    OutputControl::instance().enabled().set(APORT_FILE, false);
    OutputControl::instance().enabled().set(PCAP_FILE, false);
    OutputControl::instance().in_summary().set(PCAP_FILE, false);
    OutputControl::instance().enabled().set(PM_AGGREGATED_FILE, false);

    OutputControl::instance().enabled().set(RNC_FILE, false);

    OutputControl::instance().enabled().set("ibdiagnet2.plugin_cables", false);
    OutputControl::instance().in_summary().set("ibdiagnet2.plugin_cables", false);
    OutputControl::instance().enabled().set("csv:PLUGIN_CABLE_INFO", false);

    OutputControl::instance().enabled().set("ibdiagnet2.phy_dd_cables", false);
    OutputControl::instance().in_summary().set("ibdiagnet2.phy_dd_cables", false);
    OutputControl::instance().enabled().set("csv:PHY_DD_CABLE_INFO", false);

    OutputControl::instance().enabled().set("ibdiagnet2.phy_acc_reg_cables", false);
    OutputControl::instance().in_summary().set("ibdiagnet2.phy_acc_reg_cables", false);
    OutputControl::instance().enabled().set("csv:PHY_ACC_REG_CABLE_INFO", false);

    OutputControl::Aliases::vec_t & perf = OutputControl::instance().aliases()["perf_sections"];
    perf.push_back("csv:" SECTION_STAGE_PERF_INFO);
    perf.push_back("csv:" SECTION_CSV_PERF_INFO);
    perf.push_back("csv:" SECTION_IBIS_MAD_INFO);
    perf.push_back("csv:" SECTION_IBIS_PERF_INFO);

    IBDiagnet *app = new IBDiagnet();

    umask(0);

    int rc = 0;
    IBDiagnet::CmdLinePreprocessor cmdPreprocessor;
    rc = cmdPreprocessor.Init(argc, argv);
    if (!rc)
        rc = app->Run(cmdPreprocessor.GetArgc(), cmdPreprocessor.GetArgv());

    delete app;

    IBDIAGNET_RETURN(rc);
}

int IBDiagnet::CmdLinePreprocessor::Compare(const char *p_source, const char *p_pattern)
{
    IBDIAGNET_ENTER;

    while (*p_source == '-')
           p_source++;

    IBDIAGNET_RETURN(strcasecmp(p_source, p_pattern));
}

int IBDiagnet::CmdLinePreprocessor::Init(int argc, const char** argv)
{
    IBDIAGNET_ENTER;

    bool isCable     = false;
    bool isPhy       = false;
    bool isPCI       = false;
    bool isEnableReg = false;
    bool isDisconnetedCables = false;

    if (argc <= 0 || !argv) {
        ERR_PRINT("Wrong input parameters are provided\n");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    this->p_argv = (const char**)calloc(1, sizeof(char*) * (argc + 10));
    if (!this->p_argv)  {
        ERR_PRINT("Failed to allocate enough memory for processing the provided command line\n");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }

    memcpy(this->p_argv, argv, sizeof(char*) * argc);
    this->argumentsCount = argc;

    for (int i = 0; i < argc; i++) {
        if (!this->Compare(argv[i], "get_cable_info"))
            isCable = true;
        else if (!this->Compare(argv[i], "get_phy_info"))
            isPhy = true;
        else if (!this->Compare(argv[i], "get_p_info"))
            isPCI = true;
        else if (!this->Compare(argv[i], "enabled_reg"))
            isEnableReg = true;
        else if (!this->Compare(argv[i], "cable_info_disconnected"))
            isDisconnetedCables = true;
    }

    if (isPhy && !isEnableReg)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    if (isCable) {
        if (!isPhy)
            this->p_argv[this->argumentsCount++] = "--get_phy_info";

        this->p_argv[this->argumentsCount++] = "--enabled_reg";
        this->p_argv[this->argumentsCount++] = "dd_pddr_module,dd_pddr_lfi,pddr_module,pddr_lfi";

        if (isDisconnetedCables)
            this->p_argv[this->argumentsCount++] = "--phy_cable_disconnected";

        if (isPCI) {
            this->p_argv[this->argumentsCount++] = "--enabled_reg";
            this->p_argv[this->argumentsCount++] = "all_pci";
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

IBDiagnet::CmdLinePreprocessor::~CmdLinePreprocessor()
{
    IBDIAGNET_ENTER;

    free (this->p_argv);

    IBDIAGNET_RETURN_VOID;
}
