 /*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <math.h>
#include <iomanip>
#include "diagnostic_data.h"
#include "infiniband/ibutils/output_stream_primitives.h"

#define HILOW_TO_UINT64(x) ((uint64_t(x##_high) << 32) + x##_low)

string DiagnosticDataModuleInfo::transmitter_technology_arr[] = {
    "850 nm VCSEL"                                             ,
    "1310 nm VCSEL"                                            ,
    "1550 nm VCSEL"                                            ,
    "1310 nm FP"                                               ,
    "1310 nm DFB"                                              ,
    "1550 nm DFB"                                              ,
    "1310 nm EML"                                              ,
    "1550 nm EML"                                              ,
    "others"                                                   ,
    "1490 nm DFB"                                              ,
    "Copper cable- unequalized"                                ,
    "Copper cable- passive equalized"                          ,
    "Copper cable- near and far end limiting active equalizers",
    "Copper cable- far end limiting active equalizers"         ,
    "Copper cable- near end limiting active equalizers"        ,
    "Copper cable- linear active equalizers"                   ,
    NO_DATA_VAL };

string DiagnosticDataModuleInfo::ib_compliance_code_arr[] = {
    "SDR/",
    "DDR/",
    "QDR/",
    "FDR/",
    "EDR/",
    "HDR/",
    "NDR/" };

DiagnosticDataInfo::DiagnosticDataInfo(int page_id, int support_version,
                               int num_fields, const string &name,
                               u_int64_t not_supported_bit, u_int32_t dd_type,
                               string header, u_int8_t is_per_node,
                               PhyPluginSupportedNodesType support_nodes,
                               bool enable_disconnected_ports):
                               m_page_id(page_id),
                               m_support_version(support_version),
                               m_num_fields(num_fields),
                               m_not_supported_bit(not_supported_bit),
                               m_dd_type(dd_type),
                               m_header(header),
                               m_is_per_node(is_per_node),
                               m_support_nodes(support_nodes),
                               m_p_ibdiag(NULL),
                               m_name(name),
                               m_enable_disconnected_ports(enable_disconnected_ports)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    int status = csv_out.DumpStart(this->GetSectionHeader().c_str());
    if (status)
        IBDIAG_RETURN(status);

    sstream << "NodeGuid,PortGuid,PortNum,Version";

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataInfo::DumpDiagnosticDataHeaderEnd(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    csv_out.DumpEnd(this->GetSectionHeader().c_str());

    IBDIAGNET_RETURN_VOID;
}

bool DiagnosticDataInfo::IsDDPageSupportedInNode(struct DDPageIdentification *p_page_identification)
{
    IBDIAGNET_ENTER;

    u_int32_t group = this->m_page_id / (sizeof(p_page_identification->pages_group[0]) * 8);
    u_int32_t bit =  this->m_page_id % (sizeof(p_page_identification->pages_group[0]) * 8);

    IBDIAGNET_RETURN(p_page_identification->pages_group[group] & (1 << bit));
}


/***********************************************
 ***   Diagnostic Data Page Identification   ***
 ***********************************************/
DiagnosticDataPageIdentification::DiagnosticDataPageIdentification():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_PAGE,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_VERSION,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME,
                       NOT_SUPPORT_DD_PAGE_IDENTIFICATION,
                       DD_PHY_TYPE,
                       SECTION_PAGE_IDENTIFICATION,
                       true,
                       SUPPORT_SW)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPageIdentification::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    int status = csv_out.DumpStart(this->GetSectionHeader().c_str());
    if (status)
        IBDIAG_RETURN(status);

    sstream << "NodeGuid,Version";

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPageIdentification::DumpDiagnosticData(stringstream &sstream,
                                                          VS_DiagnosticData &dd,
                                                          IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    //unpack Diagnostic Data page identification
    struct DDPageIdentification page_identification;
    DDPageIdentification_unpack(&page_identification, (uint8_t *)&(dd.data_set));

    for (size_t i = 0;
         i < sizeof(page_identification.pages_group) / sizeof(page_identification.pages_group[0]);
         i++)
            sstream << ',' << PTR(page_identification.pages_group[i]);

    sstream << endl;
    IBDIAGNET_RETURN_VOID
}


/*****************************************
 ***   Diagnostic Data PCI functions   ***
 *****************************************/
DiagnosticDataPCI::DiagnosticDataPCI(int page_id, int support_version,
                                     int num_fields, const string &name, u_int64_t not_supported_bit,
                                     u_int32_t dd_type, string header, u_int8_t is_per_node,
                                     PhyPluginSupportedNodesType support_nodes) :
                                             DiagnosticDataInfo(page_id,
                                                                support_version,
                                                                num_fields,
                                                                name,
                                                                not_supported_bit,
                                                                dd_type,
                                                                header,
                                                                is_per_node,
                                                                support_nodes)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPCI::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    int status = csv_out.DumpStart(this->GetSectionHeader().c_str());
    if (status)
        IBDIAG_RETURN(status);

    sstream << "NodeGuid,PCIIndex,Depth,PCINode,Version";

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Diagnostic Data PCIE Counters
DiagnosticDataPCIECntrs::DiagnosticDataPCIECntrs():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_CNTRS_PAGE,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_VERSION,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_NAME,
                          NOT_SUPPORT_DIAGNOSTIC_DATA_PCIE_CNTRS,
                          DD_PCI_TYPE,
                          SECTION_PCI_CNTRS,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPCIECntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct DDPCIPerfCounters p_pci_cntrs;
    DDPCIPerfCounters_unpack(&p_pci_cntrs, (uint8_t *)&(dd.data_set));

    sstream << +p_pci_cntrs.life_time_counter_high              << ','
            << +p_pci_cntrs.life_time_counter_low               << ','
            << +p_pci_cntrs.rx_errors                           << ','
            << +p_pci_cntrs.tx_errors                           << ','
            << +p_pci_cntrs.l0_to_recovery_eieos                << ','
            << +p_pci_cntrs.l0_to_recovery_ts                   << ','
            << +p_pci_cntrs.l0_to_recovery_framing              << ','
            << +p_pci_cntrs.l0_to_recovery_retrain              << ','
            << +p_pci_cntrs.crc_error_dllp                      << ','
            << +p_pci_cntrs.crc_error_tlp                       << ','
            << +p_pci_cntrs.tx_overflow_buffer_pkt              << ','
            << +p_pci_cntrs.outbound_stalled_reads              << ','
            << +p_pci_cntrs.outbound_stalled_writes             << ','
            << +p_pci_cntrs.outbound_stalled_reads_events       << ','
            << +p_pci_cntrs.outbound_stalled_writes_events      << ','
            << +p_pci_cntrs.tx_overflow_buffer_marked_pkt       << ','
            << +p_pci_cntrs.effective_ber_magnitude             << ','
            << +p_pci_cntrs.effective_ber_coef                  << ','
            << +p_pci_cntrs.time_since_last_clear_high          << ','
            << +p_pci_cntrs.time_since_last_clear_low           ;

    IBDIAGNET_RETURN_VOID;
}

// Diagnostic Data PCIe Timers and States
DiagnosticDataPCIETimers::DiagnosticDataPCIETimers():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_TIMERS_PAGE,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_VERSION,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_NAME,
                          NOT_SUPPORT_DIAGNOSTIC_DATA_PCIE_TIMERS,
                          DD_PCI_TYPE,
                          SECTION_PCI_TIMERS,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPCIETimers::DumpDiagnosticData(stringstream &sstream,
                                                  VS_DiagnosticData &dd,
                                                  IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct DD_pcie_timers_states p_pci_timers;
    DD_pcie_timers_states_unpack(&p_pci_timers, (uint8_t *)&(dd.data_set));

    sstream << +p_pci_timers.life_time_counter_high            << ","
            << +p_pci_timers.life_time_counter_low             << ","
            << +p_pci_timers.time_to_boot_image_start          << ","
            << +p_pci_timers.time_to_link_image                << ","
            << +p_pci_timers.calibration_time                  << ","
            << +p_pci_timers.time_to_first_perst               << ","
            << +p_pci_timers.time_to_detect_state              << ","
            << +p_pci_timers.time_to_l0                        << ","
            << +p_pci_timers.time_to_crs_en                    << ","
            << +p_pci_timers.time_to_plastic_image_start       << ","
            << +p_pci_timers.time_to_iron_image_start          << ","
            << +p_pci_timers.perst_handler                     << ","
            << +p_pci_timers.times_in_l1                       << ","
            << +p_pci_timers.times_in_l23                      << ","
            << +p_pci_timers.dl_down                           << ","
            << +p_pci_timers.config_cycle1usec                 << ","
            << +p_pci_timers.config_cycle2to7usec              << ","
            << +p_pci_timers.config_cycle8to15usec             << ","
            << +p_pci_timers.config_cycle16to63usec            << ","
            << +p_pci_timers.config_cycle64usec                << ","
            << +p_pci_timers.correctable_err_msg_sent          << ","
            << +p_pci_timers.non_fatal_err_msg_sent            << ","
            << +p_pci_timers.fatal_err_msg_sent                ;

    IBDIAGNET_RETURN_VOID;
}

// Diagnostic Data PCIe Lanes Counters
DiagnosticDataPCIELanes::DiagnosticDataPCIELanes():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_LANES_PAGE,
                          DIAGNOSTIC_DATA_PCIE_LANES_VERSION,
                          DIAGNOSTIC_DATA_PCIE_LANES_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_LANES_NAME,
                          NOT_SUPPORT_DIAGNOSTIC_DATA_PCIE_LANES,
                          DD_PCI_TYPE,
                          SECTION_PCI_LANES,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPCIELanes::DumpDiagnosticData(stringstream &sstream,
                                                 VS_DiagnosticData &dd,
                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct DD_pcie_lanes_counters p_pci_lanes;
    DD_pcie_lanes_counters_unpack(&p_pci_lanes, (uint8_t *)&(dd.data_set));

    sstream << +p_pci_lanes.life_time_counter_high           << ","
            << +p_pci_lanes.life_time_counter_low            << ","
            << +p_pci_lanes.error_counter_lane0              << ","
            << +p_pci_lanes.error_counter_lane1              << ","
            << +p_pci_lanes.error_counter_lane2              << ","
            << +p_pci_lanes.error_counter_lane3              << ","
            << +p_pci_lanes.error_counter_lane4              << ","
            << +p_pci_lanes.error_counter_lane5              << ","
            << +p_pci_lanes.error_counter_lane6              << ","
            << +p_pci_lanes.error_counter_lane7              << ","
            << +p_pci_lanes.error_counter_lane8              << ","
            << +p_pci_lanes.error_counter_lane9              << ","
            << +p_pci_lanes.error_counter_lane10             << ","
            << +p_pci_lanes.error_counter_lane11             << ","
            << +p_pci_lanes.error_counter_lane12             << ","
            << +p_pci_lanes.error_counter_lane13             << ","
            << +p_pci_lanes.error_counter_lane14             << ","
            << +p_pci_lanes.error_counter_lane15             ;

    IBDIAGNET_RETURN_VOID;
}


/****************************************
 ***  Diagnostic Data PHY functions   ***
 ****************************************/
DiagnosticDataPhysLayerCntrs::DiagnosticDataPhysLayerCntrs():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_PAGE,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_VERSION,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NAME,
                       NOT_SUPPORT_DIAGNOSTIC_DATA_PHYS_CNTRS,
                       DD_PHY_TYPE,
                       SECTION_PHY_CNTRS,
                       false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPhysLayerCntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                      IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    char buffer[256];
    //unpack physical layer counters page
    struct DDPhysCounters phy_cntrs;
    DDPhysCounters_unpack(&phy_cntrs, (uint8_t *)&(dd.data_set));

    u_int32_t size = sizeof(phy_cntrs.counter) / 8;
    for (u_int32_t cnt = 0; cnt < size; cnt++) {
        snprintf(buffer, sizeof(buffer), U64H_FMT ",", phy_cntrs.counter[cnt]);
        sstream << buffer;
    }

    snprintf(buffer, sizeof(buffer), U32H_FMT "," U32H_FMT,
             phy_cntrs.counter24,
             phy_cntrs.counter25);

    sstream << buffer;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Troubleshooting Info
DiagnosticDataTroubleshootingInfo::DiagnosticDataTroubleshootingInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_PAGE,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_VERSION,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NAME,
                       NOT_SUPPORT_DD_TROUBLESHOOTING_INFO,
                       DD_PHY_TYPE,
                       SECTION_TROUBLESHOOTING_INFO,
                       false)
{
}

void DiagnosticDataTroubleshootingInfo::DumpDiagnosticData(stringstream &sstream,
                                        VS_DiagnosticData &dd, IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DDTroubleShootingInfo troubleshooting_info;
    DDTroubleShootingInfo_unpack(&troubleshooting_info, (uint8_t *)&(dd.data_set));

    sstream << troubleshooting_info.status_opcode << ','
            << '\"' << troubleshooting_info.status_message << '\"';

    IBDIAGNET_RETURN_VOID;
}

// PDDR Operation Info
DiagnosticDataOperationInfo::DiagnosticDataOperationInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_OPERATION_INFO_PAGE,
                       DIAGNOSTIC_DATA_OPERATION_INFO_VERSION,
                       DIAGNOSTIC_DATA_OPERATION_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_OPERATION_INFO_NAME,
                       NOT_SUPPORT_DD_OPERATION_INFO,
                       DD_PHY_TYPE,
                       SECTION_OPERATION_INFO,
                       false)
{
}

void DiagnosticDataOperationInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                     IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DDOperationInfo operation_info;
    DDOperationInfo_unpack(&operation_info, (uint8_t *)&(dd.data_set));

    sstream << +operation_info.proto_active                   << ','
            << +operation_info.neg_mode_active                << ','
            << +operation_info.pd_fsm_state                   << ','
            << +operation_info.phy_mngr_fsm_state             << ','
            << +operation_info.eth_an_fsm_state               << ','
            << +operation_info.ib_phy_fsm_state               << ','
            << +operation_info.phy_hst_fsm_state              << ','
            << PTR(operation_info.phy_manager_link_enabled)   << ','
            << PTR(operation_info.core_to_phy_link_enabled)   << ','
            << +operation_info.cable_proto_cap                << ','
            << PTR(operation_info.link_active)                << ','
            << +operation_info.retran_mode_active             << ','
            << +operation_info.retran_mode_request            << ','
            << +operation_info.loopback_mode                  << ','
            << +operation_info.fec_mode_active                << ','
            << +operation_info.fec_mode_request               << ','
            << +operation_info.profile_fec_in_use             << ','
            << +operation_info.eth_25g_50g_fec_support        << ','
            << +operation_info.eth_100g_fec_support           << ','
            << +operation_info.pd_link_enabled                << ','
            << +operation_info.phy_hst_link_enabled           << ','
            << +operation_info.eth_an_link_enabled            << ','
            << +operation_info.psi_fsm_state                  << ','
            << +operation_info.cable_proto_cap_ext            ;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Phy Info
DiagnosticDataPhyInfo::DiagnosticDataPhyInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_INFO_PAGE,
                       DIAGNOSTIC_DATA_PHY_INFO_VERSION,
                       DIAGNOSTIC_DATA_PHY_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHY_INFO_NAME,
                       NOT_SUPPORT_DD_PHY_INFO,
                       DD_PHY_TYPE,
                       SECTION_PHY_INFO,
                       false)
{
}

void DiagnosticDataPhyInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                               IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DDPhyInfo phy_info;
    DDPhyInfo_unpack(&phy_info, (uint8_t *)&(dd.data_set));

    sstream << +phy_info.remote_device_type << ','
            << +phy_info.port_notifications << ','
            << phy_info.num_of_negotiation_attempts << ','
            << +phy_info.ib_revision << ','
            << +phy_info.lp_ib_revision << ','
            << +phy_info.hw_link_phy_state << ','
            << phy_info.phy_manger_disable_mask << ','
            << phy_info.pcs_phy_state << ','
            << phy_info.lp_proto_enabled << ','
            << phy_info.lp_fec_mode_support << ','
            << phy_info.lp_fec_mode_request << ','
            << phy_info.ib_last_link_down_reason << ','
            << phy_info.speed_deg_db << ','
            << phy_info.degrade_grade_lane_0 << ','
            << phy_info.degrade_grade_lane_1 << ','
            << phy_info.degrade_grade_lane_2 << ','
            << phy_info.degrade_grade_lane_3 << ','
            << phy_info.ib_phy_fsm_state_trace << ','
            << +phy_info.irisc_status << ','
            << phy_info.stamping_reason;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Module Info
DiagnosticDataModuleInfo::DiagnosticDataModuleInfo(bool enable_disconnected_ports):
    DiagnosticDataInfo(DIAGNOSTIC_DATA_MODULE_INFO_PAGE,
                       DIAGNOSTIC_DATA_MODULE_INFO_VERSION,
                       DIAGNOSTIC_DATA_MODULE_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_MODULE_INFO_NAME,
                       NOT_SUPPORT_DD_MODULE_INFO,
                       DD_PHY_TYPE,
                       SECTION_MODULE_INFO,
                       false,
                       SUPPORT_SW,
                       enable_disconnected_ports)
{
}

void DiagnosticDataModuleInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                  IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct DDModuleInfo module_info;
    DDModuleInfo_unpack(&module_info, (uint8_t *)&(dd.data_set));

    DumpModuleInfoRawData(sstream, module_info);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataModuleInfo::DumpModuleInfoRawData(stringstream &sstream, const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    uint8_t fw_hi = (uint8_t)(module_info.fw_version >> 24);
    uint8_t fw_mid = (uint8_t)((module_info.fw_version >> 16) & 0x00ff);
    uint16_t fw_lo = (uint16_t)(module_info.fw_version & 0x0000ffff);

    sstream << +module_info.cable_technology                  << ','
            << +module_info.cable_breakout                    << ','
            << +module_info.ext_ethernet_compliance_code      << ','
            << +module_info.ethernet_compliance_code          << ','
            << +module_info.cable_type                        << ','
            << +module_info.cable_vendor                      << ','
            << +module_info.cable_length                      << ','
            << +module_info.cable_identifier                  << ','
            << +module_info.cable_power_class                 << ','
            << +module_info.max_power                         << ','
            << +module_info.cable_rx_amp                      << ','
            << +module_info.cable_rx_emphasis                 << ','
            << +module_info.cable_tx_equalization             << ','
            << +module_info.cable_attenuation_25g             << ','
            << +module_info.cable_attenuation_12g             << ','
            << +module_info.cable_attenuation_7g              << ','
            << +module_info.cable_attenuation_5g              << ','
            << +module_info.cable_rx_post_emphasis            << ','
            << +module_info.rx_cdr_cap                        << ','
            << +module_info.tx_cdr_cap                        << ','
            << +module_info.rx_cdr_state                      << ','
            << +module_info.tx_cdr_state                      << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_name)) << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_pn))   << ','
            << QUOTED(DescToCsvDesc(ConvertRevisionToStr(module_info))) << ','
            << +fw_hi << "." << +fw_mid  << "." << fw_lo      << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_sn))   << ','
            << +module_info.temperature                       << ','
            << +module_info.voltage                           << ','
            << +module_info.rx_power_lane0                    << ','
            << +module_info.rx_power_lane1                    << ','
            << +module_info.rx_power_lane2                    << ','
            << +module_info.rx_power_lane3                    << ','
            << +module_info.rx_power_lane4                    << ','
            << +module_info.rx_power_lane5                    << ','
            << +module_info.rx_power_lane6                    << ','
            << +module_info.rx_power_lane7                    << ','
            << +module_info.tx_power_lane0                    << ','
            << +module_info.tx_power_lane1                    << ','
            << +module_info.tx_power_lane2                    << ','
            << +module_info.tx_power_lane3                    << ','
            << +module_info.tx_power_lane4                    << ','
            << +module_info.tx_power_lane5                    << ','
            << +module_info.tx_power_lane6                    << ','
            << +module_info.tx_power_lane7                    << ','
            << +module_info.tx_bias_lane0                     << ','
            << +module_info.tx_bias_lane1                     << ','
            << +module_info.tx_bias_lane2                     << ','
            << +module_info.tx_bias_lane3                     << ','
            << +module_info.tx_bias_lane4                     << ','
            << +module_info.tx_bias_lane5                     << ','
            << +module_info.tx_bias_lane6                     << ','
            << +module_info.tx_bias_lane7                     << ','
            << +module_info.temperature_high_th               << ','
            << +module_info.temperature_low_th                << ','
            << +module_info.voltage_high_th                   << ','
            << +module_info.voltage_low_th                    << ','
            << +module_info.rx_power_high_th                  << ','
            << +module_info.rx_power_low_th                   << ','
            << +module_info.tx_power_high_th                  << ','
            << +module_info.tx_power_low_th                   << ','
            << +module_info.tx_bias_high_th                   << ','
            << +module_info.tx_bias_low_th                    << ','
            << +module_info.module_st                         << ','
            << +module_info.rx_power_type                     << ','
            << +module_info.did_cap                           << ','
            << +module_info.smf_length                        << ','
            << +module_info.wavelength                        << ','
            << +module_info.active_set_host_compliance_code   << ','
            << +module_info.active_set_media_compliance_code  << ','
            << +module_info.ib_compliance_code                << ','
            << +module_info.nbr_250                           << ','
            << +module_info.nbr_100                           << ','
            << +module_info.monitor_cap_mask                  << ','
            << +module_info.ib_width                          << ','
            << +module_info.dp_st_lane0                       << ','
            << +module_info.dp_st_lane1                       << ','
            << +module_info.dp_st_lane2                       << ','
            << +module_info.dp_st_lane3                       << ','
            << +module_info.dp_st_lane4                       << ','
            << +module_info.dp_st_lane5                       << ','
            << +module_info.dp_st_lane6                       << ','
            << +module_info.dp_st_lane7                       << ','
            << +module_info.length_om2                        << ','
            << +module_info.length_om3                        << ','
            << +module_info.length_om4                        << ','
            << +module_info.length_om5                        << ','
            << +module_info.length_om1                        << ','
            << +module_info.wavelength_tolerance              << ','
            << +module_info.memory_map_rev                    << ','
            << +module_info.memory_map_compliance             << ','
            << "0x" << HEX(module_info.date_code)             << ','
            << +module_info.connector_type                    << ','
            << +module_info.vendor_oui                        << ','
            << +module_info.rx_input_valid                    << ','
            << +module_info.rx_output_valid                   << ','
            << +module_info.tx_input_freq_sync                << ','
            << +module_info.error_code                        ;

    IBDIAGNET_RETURN_VOID;
}

const string& DiagnosticDataModuleInfo::ConvertTransmitterTechnologyToStr(
                                                         u_int8_t transmitter_technology)
{
    IBDIAGNET_ENTER;

    u_int8_t size = (u_int8_t)ARRAY_SIZE(DiagnosticDataModuleInfo::transmitter_technology_arr);
    if (transmitter_technology >= size)
        IBDIAGNET_RETURN(DiagnosticDataModuleInfo::transmitter_technology_arr[size - 1]);

    IBDIAGNET_RETURN(DiagnosticDataModuleInfo::transmitter_technology_arr[transmitter_technology]);
}

#ifndef BIT_TO_STR
    #define BIT_TO_STR(val, index) (val & (1 << index) ? "1" : "0")
#else
    #error "the BIT_TO_STR has already been defined"
#endif

string DiagnosticDataModuleInfo::ConvertCableTechnologyBitsToStr(
                                                             const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    stringstream cable_technology_ss;
    if (DiagnosticDataModuleInfo::IsCMISCable(module_info.cable_identifier))
        cable_technology_ss << NO_DATA_VAL << ',' << NO_DATA_VAL << ','
                            << NO_DATA_VAL << ',' << NO_DATA_VAL;
    else
        cable_technology_ss << BIT_TO_STR(module_info.cable_technology, 0) << ','
                            << BIT_TO_STR(module_info.cable_technology, 1) << ','
                            << BIT_TO_STR(module_info.cable_technology, 2) << ','
                            << BIT_TO_STR(module_info.cable_technology, 3);

    IBDIAGNET_RETURN(cable_technology_ss.str());
}

string DiagnosticDataModuleInfo::ConvertCableIdentifierToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    string result;

    switch (module_info.cable_identifier) {
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP28:
        result = "QSFP28";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_PLUS:
        result = "QSFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus:
        result = "SFP28/SFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSA:
        result = "QSA (QSFP->SFP)";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_Backplane:
        result = "Backplane";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
        result = "SFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
        result = "QSFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
        result = "QSFP_CMIS";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
        result = "OSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_C2C:
        result = "C2C";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
        result = "DSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_SPLIT_CABLE:
        result = "QSFP_Split_Cable";
        break;
    default:
        result = NO_DATA_VAL;
        break;
    }

    IBDIAGNET_RETURN(result);
}

string DiagnosticDataModuleInfo::ConvertCableTemperatureToStr(const struct DDModuleInfo &module_info,
                                                              u_int16_t temp)
{
    IBDIAGNET_ENTER;

    stringstream temperature_ss;
    u_int8_t cable_technology = module_info.cable_technology >> 4;
    int8_t temperature = (int8_t)(temp >> 8);
    if (cable_technology == CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED ||
            cable_technology == CABLE_MODULE_INFO_TYPE_COPPER_PASSIVE_EQUALIZED ||
            temperature < CABLE_MODULE_INFO_MIN_SUPPORT_TEMPERATURE ||
            temperature > CABLE_MODULE_INFO_MAX_SUPPORT_TEMPERATURE) {
        temperature_ss << NO_DATA_VAL;
        IBDIAGNET_RETURN(temperature_ss.str());
    }

    temperature_ss << +temperature << 'C';

    IBDIAGNET_RETURN(temperature_ss.str());
}

string DiagnosticDataModuleInfo::ConvertVoltageToStr(u_int16_t vcc)
{
    IBDIAGNET_ENTER;

    stringstream voltage_ss;
    voltage_ss << vcc * 100 << "uV";

    IBDIAGNET_RETURN(voltage_ss.str());
}

string DiagnosticDataModuleInfo::ConvertMaxPowerToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if (!module_info.max_power)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream max_power_ss;
    max_power_ss << (double)module_info.max_power / 4.0 << " W";
    IBDIAGNET_RETURN(max_power_ss.str());
}

bool DiagnosticDataModuleInfo::IsCMISCable(uint8_t identifier)
{
    IBDIAGNET_ENTER;

    switch (identifier)
    {
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
          IBDIAGNET_RETURN(true);

    default:
          IBDIAGNET_RETURN(false);
    }
}

string DiagnosticDataModuleInfo::ConvertCableLengthToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if (!module_info.cable_length)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream cable_length_ss;
    u_int64_t cable_length = (u_int64_t)module_info.cable_length;
    if (DiagnosticDataModuleInfo::IsCMISCable(module_info.cable_identifier)) {
        u_int8_t cable_length_value = module_info.cable_length & 0x3f;
        if (!cable_length_value)
            IBDIAGNET_RETURN(NO_DATA_VAL);

        u_int8_t cable_length_mul = module_info.cable_length >> 6;
        u_int64_t mul = 0;
        if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_0_1) {
            double cable_length_float = 0.1 * (double)cable_length_value;
            cable_length_ss << cable_length_float << " m";
            IBDIAGNET_RETURN(cable_length_ss.str());
        } else {
            if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_1)
                mul = 1;
            else if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_10)
                mul = 10;
            else if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_100)
                mul = 100;
        }

        cable_length = mul * (u_int64_t)cable_length_value;
    }

    cable_length_ss << cable_length << " m";

    IBDIAGNET_RETURN(cable_length_ss.str());
}

string DiagnosticDataModuleInfo::ConvertCableLengthSMFiberToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    u_int16_t length_base = module_info.smf_length & 0xff;
    u_int16_t unit_bit = (module_info.smf_length >> 8) & 0x3;
    stringstream smf_length_ss;

    if (!length_base)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    switch (unit_bit) {
    case 0:
        smf_length_ss << length_base << " km";
        break;
    case 0x1:
        smf_length_ss << (double)(length_base * 100) / 1000.0 << " km";
        break;
    default:
        smf_length_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(smf_length_ss.str());
}

string DiagnosticDataModuleInfo::ConvertCableLengthOMXToStr(const struct DDModuleInfo &module_info,
                                                            u_int8_t om_index)
{
    IBDIAGNET_ENTER;

    u_int8_t unit;
    u_int8_t length_omx;
    switch (om_index) {
    case 1:
        length_omx = module_info.length_om1;
        unit = 10;
        break;
    case 2:
        length_omx = module_info.length_om2;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
            unit = 10;
        else
            unit = 1;
        break;
    case 3:
        length_omx = module_info.length_om3;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 4:
        length_omx = module_info.length_om4;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 5:
        length_omx = module_info.length_om5;
        unit = 2;
        break;
    default:
        IBDIAGNET_RETURN(NO_DATA_VAL);
    }

    if (!length_omx)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream length_omx_ss;
    length_omx_ss << (u_int64_t)length_omx * unit << " m";
    IBDIAGNET_RETURN(length_omx_ss.str());
}

string DiagnosticDataModuleInfo::ConvertIBComplianceCodeToStr(u_int8_t ib_compliance_code)
{
    IBDIAGNET_ENTER;
    stringstream ib_compliance_code_ss;

    u_int8_t size = ARRAY_SIZE(DiagnosticDataModuleInfo::ib_compliance_code_arr);
    for (u_int8_t i = 0; i < size; i++) {
         if (ib_compliance_code & (1 << i))
             ib_compliance_code_ss << DiagnosticDataModuleInfo::ib_compliance_code_arr[i];
    }

    string ib_compliance_code_str = ib_compliance_code_ss.str();
    if (ib_compliance_code_str.empty()) {
        ib_compliance_code_str = NO_DATA_VAL;
    } else {
        ib_compliance_code_str.erase(ib_compliance_code_str.size() - 1, 1);
    }

    IBDIAGNET_RETURN(ib_compliance_code_str);
}

u_int8_t DiagnosticDataModuleInfo::SelectNominalBR(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if (module_info.nbr_100 == 0xff)
        IBDIAGNET_RETURN(module_info.nbr_250 / 4);

    IBDIAGNET_RETURN(module_info.nbr_100 / 10);
}

u_int8_t DiagnosticDataModuleInfo::SelectTransmitterTechnology(const struct DDModuleInfo &module_info)
{
     IBDIAGNET_ENTER;

     if (DiagnosticDataModuleInfo::IsCMISCable(module_info.cable_identifier))
         IBDIAGNET_RETURN(module_info.cable_technology);

     IBDIAGNET_RETURN(module_info.cable_technology >> 4);
}

bool DiagnosticDataModuleInfo::IsModule(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if (((module_info.cable_technology >> 4) != CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED) &&
        (module_info.connector_type != CABLE_MODULE_INFO_NO_SEPARABLE_CONNECTOR))
        IBDIAGNET_RETURN (true);
    IBDIAGNET_RETURN(false);
}

bool DiagnosticDataModuleInfo::IsActiveCable(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if (((module_info.cable_technology >> 4) != CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED) &&
        (module_info.connector_type == CABLE_MODULE_INFO_NO_SEPARABLE_CONNECTOR))
        IBDIAGNET_RETURN(true);
    IBDIAGNET_RETURN(false);
}

bool DiagnosticDataModuleInfo::IsPassiveCable(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    if ((module_info.cable_technology >> 4) == CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED)
        IBDIAGNET_RETURN(true);
    IBDIAGNET_RETURN(false);
}

string DiagnosticDataModuleInfo::ConvertCDREnableTxRxToStr(const struct DDModuleInfo &module_info,
                                                           bool is_rx,
                                                           bool is_csv)
{
    IBDIAGNET_ENTER;

    stringstream cdr_state_ss;

    if (DiagnosticDataModuleInfo::IsModule(module_info) ||
        DiagnosticDataModuleInfo::IsActiveCable(module_info)) {
        u_int8_t cdr_cap;
        u_int8_t cdr_state;
        if (is_rx) {
            cdr_cap = module_info.rx_cdr_cap;
            cdr_state = module_info.rx_cdr_state;
        } else {
            cdr_cap = module_info.tx_cdr_cap;
            cdr_state = module_info.tx_cdr_state;
        }
        if (is_csv || cdr_cap) {
            cdr_state_ss << PTR(cdr_state);
        } else {
            cdr_state_ss << NO_DATA_VAL;
        }
    } else {
        cdr_state_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(cdr_state_ss.str());
}

string DiagnosticDataModuleInfo::ConvertTxEQRxAMPRxEMPToStr(const struct DDModuleInfo &module_info,
                                                            u_int8_t val,
                                                            bool is_csv)
{
    IBDIAGNET_ENTER;

    stringstream val_ss;
    if (DiagnosticDataModuleInfo::IsModule(module_info) ||
        DiagnosticDataModuleInfo::IsActiveCable(module_info)) {
        if (!is_csv)
            val_ss << +val << " " << +val << " " << +val << " " << +val;
        else
            val_ss << HEX(val) << HEX(val) << HEX(val) << HEX(val);
    } else {
        val_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(val_ss.str());
}

string DiagnosticDataModuleInfo::ConvertAttenuationToStr(const struct DDModuleInfo &module_info,
                                                         bool is_csv)
{
    IBDIAGNET_ENTER;
    stringstream attenuation_ss;

    string separator;
    string na_str;
    if (is_csv) {
        separator = ",";
    } else {
        separator = " ";
    }

    if (DiagnosticDataModuleInfo::IsPassiveCable(module_info)) {
        attenuation_ss << +module_info.cable_attenuation_5g << separator
                       << +module_info.cable_attenuation_7g << separator
                       << +module_info.cable_attenuation_12g << separator
                       << +module_info.cable_attenuation_25g;
    } else {
        attenuation_ss << NO_DATA_VAL << separator << NO_DATA_VAL << separator << NO_DATA_VAL
                       << separator << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(attenuation_ss.str());
}

string DiagnosticDataModuleInfo::ConvertFWVersionToStr(const struct DDModuleInfo &module_info,
                                                       bool is_csv)
{
    IBDIAGNET_ENTER;
    stringstream fw_version_ss;

    if ((module_info.fw_version != 0) &&
        (DiagnosticDataModuleInfo::IsModule(module_info) ||
         DiagnosticDataModuleInfo::IsActiveCable(module_info))) {
        uint8_t fw_hi = (uint8_t)(module_info.fw_version >> 24);
        uint8_t fw_mid = (uint8_t)((module_info.fw_version >> 16) & 0x00ff);
        uint16_t fw_lo = (uint16_t)(module_info.fw_version & 0x0000ffff);
        fw_version_ss << +fw_hi << "." << +fw_mid << "." << fw_lo;
    } else  {
        fw_version_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(fw_version_ss.str());
}

string DiagnosticDataModuleInfo::ConvertDateCodeToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    u_int64_t date_code = module_info.date_code >> 16;
    char *p_date_code = (char *)&date_code;
    for (int i = 0; i < 6; i++) {
         if (!isdigit(p_date_code[i]))
             IBDIAGNET_RETURN(NO_DATA_VAL);
    }

    stringstream date_code_ss;
    date_code_ss << p_date_code[5] << p_date_code[4] << '-'
                 << p_date_code[3] << p_date_code[2] << '-'
                 << p_date_code[1] << p_date_code[0];

    IBDIAGNET_RETURN(date_code_ss.str());
}

string DiagnosticDataModuleInfo::ConvertRevisionToStr(const struct DDModuleInfo &module_info)
{
    IBDIAGNET_ENTER;

    stringstream rev_ss;
    size_t size = sizeof(module_info.vendor_rev);
    for (size_t i = 0; i < size; i++) {
         if (module_info.vendor_rev[i])
             rev_ss << module_info.vendor_rev[i];
    }

    IBDIAGNET_RETURN(rev_ss.str());
}

double DiagnosticDataModuleInfo::MWToDBM(double mW)
{
    if (mW == 0)
        IBDIAGNET_RETURN(-999.999);
    IBDIAGNET_RETURN(10 * log10(mW));
}

void DiagnosticDataModuleInfo::DumpModuleInfoHeader(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "Vendor,OUI,PN,SN,REV,Identifier,IdentifierStr,"
            << "LengthCopperOrActive,LengthSMFiber,LengthOM1,LengthOM2,LengthOM3,LengthOM4,"
            << "LengthOM5,TypeDesc,Temperature,PowerClass,MaxPower,Connector,Type,SupportedSpeed,"
            << "SupportedSpeedDesc,NominalBitrate,CDREnableRx,CDREnableTx,"
            << "InputEq,OutputAmp,OutputPreEmp,OutputPostEmp,"
            << "FWVersion,Attenuation5G,Attenuation7G,Attenuation12G,Attenuation25G,RXPowerType,"
            << "module_info_ext,RX1Power,RX2Power,RX3Power,RX4Power,"
            << "TX1Bias,TX2Bias,TX3Bias,TX4Bias,"
            << "TX1Power,TX2Power,TX3Power,TX4Power,SupplyVoltageReporting,"
            << "DateCode,Lot,TransmitterTechnology,ActiveWavelengthControl,CooledTransmitterDevice,"
            << "ActivePinDetector,TunableTransmitter,ExtendedSpecificationComplianceCodes,"
            << "AlarmTemperatureHighThresh,AlarmTemperatureLowThresh,"
            << "WarnTemperatureHighThresh,WarnTemperatureLowThresh,"
            << "AlarmVoltageHighThresh,AlarmVoltageLowThresh,"
            << "WarnVoltageHighThresh,WarnVoltageLowThresh,"
            << "RXPowerHighThresh,RXPowerLowThresh,TXPowerHighThresh,TXPowerLowThresh,"
            << "TXBiasHighThresh,TXBiasLowThresh";

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataModuleInfo::DumpModuleInfoData(stringstream &sstream,
                                                  VS_DiagnosticData *dd)
{
    IBDIAGNET_ENTER;

    struct DDModuleInfo module_info;
    struct DDModuleInfo *p_module_info = NULL;

    if (dd)
        DDModuleInfo_unpack(p_module_info = &module_info, (uint8_t *)&(dd->data_set));

    DumpModuleInfoData(sstream, p_module_info, 0);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataModuleInfo::DumpModuleInfoData(stringstream &sstream,
                                                  const struct DDModuleInfo *p_module_info,
                                                  u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info) {
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A";
        IBDIAGNET_RETURN_VOID;
    }

    u_int8_t transmitter_technology =
                               DiagnosticDataModuleInfo::SelectTransmitterTechnology(*p_module_info);

    sstream << QUOTED(DescToCsvDesc(p_module_info->vendor_name))                                 << ','
            << PTR(p_module_info->vendor_oui)                                                    << ','
            << QUOTED(DescToCsvDesc(p_module_info->vendor_pn))                                   << ','
            << QUOTED(DescToCsvDesc(p_module_info->vendor_sn))                                   << ','
            << QUOTED(DescToCsvDesc(ConvertRevisionToStr(*p_module_info)))                       << ','
            << PTR(p_module_info->cable_identifier)                                              << ','
            << QUOTED(ConvertCableIdentifierToStr(*p_module_info))                               << ','
            << QUOTED(ConvertCableLengthToStr(*p_module_info))                                   << ','
            << QUOTED(ConvertCableLengthSMFiberToStr(*p_module_info))                            << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 1))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 2))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 3))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 4))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 5))                             << ','
            << QUOTED(ConvertTransmitterTechnologyToStr(transmitter_technology))                 << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature))  << ','
            << +p_module_info->cable_power_class                                                 << ','
            << ConvertMaxPowerToStr(*p_module_info)                                              << ','
            << PTR(p_module_info->connector_type)                                                << ','
            << +transmitter_technology                                                           << ','
            << +p_module_info->ib_compliance_code                                                << ','
            << QUOTED(ConvertIBComplianceCodeToStr(p_module_info->ib_compliance_code))           << ','
            << (SelectNominalBR(*p_module_info) * 1000)                                          << ','
            << ConvertCDREnableTxRxToStr(*p_module_info, true, true)                             << ','
            << ConvertCDREnableTxRxToStr(*p_module_info, false, true)                            << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_tx_equalization, true)            << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info, p_module_info->cable_rx_amp, true)     << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_rx_emphasis, true)                << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_rx_post_emphasis, true)           << ','
            << ConvertFWVersionToStr(*p_module_info, true)                                       << ','
            << ConvertAttenuationToStr(*p_module_info, true)                                     << ','
            //module_info_ext is not available here
            << +p_module_info->rx_power_type                                                     << ','
            << module_info_ext                                                                   << ','
            << (float)(p_module_info->rx_power_lane0)                                            << ','
            << (float)(p_module_info->rx_power_lane1)                                            << ','
            << (float)(p_module_info->rx_power_lane2)                                            << ','
            << (float)(p_module_info->rx_power_lane3)                                            << ','
            << (float)(p_module_info->tx_bias_lane0 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane1 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane2 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane3 / 500.0)                                     << ','
            << (float)(p_module_info->tx_power_lane0)                                            << ','
            << (float)(p_module_info->tx_power_lane1)                                            << ','
            << (float)(p_module_info->tx_power_lane2)                                            << ','
            << (float)(p_module_info->tx_power_lane3 )                                           << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage))                               << ','
            << ConvertDateCodeToStr(*p_module_info)                                              << ','
            << PTR((u_int16_t)(p_module_info->date_code & 0xffff))                               << ','
            << PTR(transmitter_technology)                                                       << ','
            << ConvertCableTechnologyBitsToStr(*p_module_info)                                   << ','
            << +p_module_info->ext_ethernet_compliance_code                                      << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info,
                                                   p_module_info->temperature_high_th))          << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info,
                                                   p_module_info->temperature_low_th))           << ','
            << "N/A"                                                                             << ','
            << "N/A"                                                                             << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage_high_th))                       << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage_low_th))                        << ','
            << "N/A"                                                                             << ','
            << "N/A"                                                                             << ','
            << +p_module_info->rx_power_high_th                                                  << ','
            << +p_module_info->rx_power_low_th                                                   << ','
            << +p_module_info->tx_power_high_th                                                  << ','
            << +p_module_info->tx_power_low_th                                                   << ','
            << +p_module_info->tx_bias_high_th                                                   << ','
            << +p_module_info->tx_bias_low_th                                                    ;

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataModuleInfo::DumpModuleInfoData(ofstream &sout,
                                                  VS_DiagnosticData *dd)
{
    IBDIAGNET_ENTER;

    struct DDModuleInfo module_info;
    struct DDModuleInfo *p_module_info = NULL;

    if (dd)
        DDModuleInfo_unpack(p_module_info = &module_info, (uint8_t *)&(dd->data_set));

    DumpModuleInfoData(sout, p_module_info, 0);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataModuleInfo::DumpModuleInfoData(ofstream &sout,
                                                  const struct DDModuleInfo *p_module_info,
                                                  u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info) {
         sout << "Identifier: N/A" << endl
              << "Vendor: N/A" << endl
              << "OUI: N/A" << endl
              << "PN: N/A"  << endl
              << "SN: N/A"  << endl
              << "Rev: N/A"   << endl
              << "LengthCopperOrActive: N/A" << endl
              << "LengthSMFiber: N/A" << endl
              << "LengthOM1: N/A" << endl
              << "LengthOM2: N/A" << endl
              << "LengthOM3: N/A" << endl
              << "LengthOM4: N/A" << endl
              << "LengthOM5: N/A" << endl
              << "Type: N/A" << endl
              << "SupportedSpeed: N/A" << endl
              << "Temperature: N/A"  << endl
              << "PowerClass: N/A" << endl
              << "NominalBitrate: N/A"  << endl
              << "CDREnableRx: N/A"  << endl
              << "CDREnableTx: N/A"  << endl
              << "InputEq: N/A"  << endl
              << "OutputAmp: N/A"  << endl
              << "OutputPreEmp: N/A"  << endl
              << "OutputPostEmp: N/A" << endl
              << "FW Version: N/A" <<  endl
              << "Attenuation(5,7,12,25): N/A" << endl
              << "RX power type: N/A" << endl
              << "RX1 Power: N/A" << endl
              << "RX2 Power: N/A" << endl
              << "RX3 Power: N/A" << endl
              << "RX4 Power: N/A" << endl
              << "TX1 Bias: N/A" << endl
              << "TX2 Bias: N/A" << endl
              << "TX3 Bias: N/A" << endl
              << "TX4 Bias: N/A" << endl
              << "TX1 Power: N/A" << endl
              << "TX2 Power: N/A" << endl
              << "TX3 Power: N/A" << endl
              << "TX4 Power: N/A" << endl
              << "Supply Voltage Reporting: N/A" << endl
              << "DateCode: N/A" << endl
              << "Lot: N/A" << endl
              << "Transmitter Technology: N/A" << endl
              << "Extended Specification Compliance Codes: N/A" << endl
              << "AlarmTemperatureHighThresh: N/A" << endl
              << "AlarmTemperatureLowThresh: N/A" << endl
              << "WarnTemperatureHighThresh: N/A" << endl
              << "WarnTemperatureLowThresh: N/A" << endl
              << "AlarmVoltageHighThresh: N/A" << endl
              << "AlarmVoltageLowThresh: N/A" << endl
              << "WarnVoltageHighThresh: N/A" << endl
              << "WarnVoltageLowThresh: N/A";
         IBDIAGNET_RETURN_VOID;
    }

    u_int8_t transmitter_technology =
                              DiagnosticDataModuleInfo::SelectTransmitterTechnology(*p_module_info);
    sout << "Identifier: " << ConvertCableIdentifierToStr(*p_module_info) << " ("
            << PTR(p_module_info->cable_identifier) << ')' << endl
         << "Vendor: " << DescToCsvDesc(p_module_info->vendor_name) << endl
         << "OUI: " << PTR(p_module_info->vendor_oui) << endl
         << "PN: " << DescToCsvDesc(p_module_info->vendor_pn) << endl
         << "SN: " << DescToCsvDesc(p_module_info->vendor_sn) << endl
         << "Rev: " << DescToCsvDesc(ConvertRevisionToStr(*p_module_info)) << endl
         << "LengthCopperOrActive: " << ConvertCableLengthToStr(*p_module_info) << endl
         << "LengthSMFiber: " << ConvertCableLengthSMFiberToStr(*p_module_info) << endl
         << "LengthOM1: " << ConvertCableLengthOMXToStr(*p_module_info, 1) << endl
         << "LengthOM2: " << ConvertCableLengthOMXToStr(*p_module_info, 2) << endl
         << "LengthOM3: " << ConvertCableLengthOMXToStr(*p_module_info, 3) << endl
         << "LengthOM4: " << ConvertCableLengthOMXToStr(*p_module_info, 4) << endl
         << "LengthOM5: " << ConvertCableLengthOMXToStr(*p_module_info, 5) << endl
         << "Type: "
         << ConvertTransmitterTechnologyToStr(transmitter_technology) << endl
         << "SupportedSpeed: "
         << ConvertIBComplianceCodeToStr(p_module_info->ib_compliance_code) << endl
         << "Temperature: " << ConvertCableTemperatureToStr(*p_module_info,
                                                            p_module_info->temperature) << endl
         << "PowerClass: " << +p_module_info->cable_power_class << endl
         << "NominalBitrate: " << +SelectNominalBR(*p_module_info) << endl
         << "CDREnableRx: " << ConvertCDREnableTxRxToStr(*p_module_info, true, false) << endl
         << "CDREnableTx: " << ConvertCDREnableTxRxToStr(*p_module_info, false, false) << endl
         << "InputEq: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                      p_module_info->cable_tx_equalization,
                                                      false) << endl
         << "OutputAmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                        p_module_info->cable_rx_amp,
                                                        false) << endl
         << "OutputPreEmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                           p_module_info->cable_rx_emphasis,
                                                           false) << endl
         << "OutputPostEmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                            p_module_info->cable_rx_post_emphasis,
                                                            false) << endl
         << "FW Version: " << ConvertFWVersionToStr(*p_module_info, false) << endl
         << "Attenuation(5,7,12,25): " << ConvertAttenuationToStr(*p_module_info, false) << endl
         << "RX power type: " << (p_module_info->rx_power_type ? "Average Power" : "OMA") << endl
         << "RX1 Power: " << p_module_info->rx_power_lane0 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX2 Power: " << p_module_info->rx_power_lane1 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX3 Power: " << p_module_info->rx_power_lane2 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX4 Power: " << p_module_info->rx_power_lane3 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX1 Bias: " << (float)(p_module_info->tx_bias_lane0 / 500.0) << " mA" << endl
         << "TX2 Bias: " << (float)(p_module_info->tx_bias_lane1 / 500.0) << " mA" << endl
         << "TX3 Bias: " << (float)(p_module_info->tx_bias_lane2 / 500.0) << " mA" << endl
         << "TX4 Bias: " << (float)(p_module_info->tx_bias_lane3 / 500.0) << " mA" << endl
         << "TX1 Power: " << p_module_info->tx_power_lane0 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX2 Power: " << p_module_info->tx_power_lane1 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX3 Power: " << p_module_info->tx_power_lane2 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX4 Power: " << p_module_info->tx_power_lane3 << (module_info_ext ? " uW" : " dBm") << endl
         << "Supply Voltage Reporting: " << ConvertVoltageToStr(p_module_info->voltage) << endl
         << "DateCode: "<< ConvertDateCodeToStr(*p_module_info) << endl
         << "Lot: " << PTR((u_int16_t)(p_module_info->date_code & 0xffff))<< endl
         << "Transmitter Technology: " << PTR(p_module_info->cable_technology) << endl
         << "Extended Specification Compliance Codes: "
            << +p_module_info->ext_ethernet_compliance_code << endl
         << "AlarmTemperatureHighThresh: "
            << ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature_high_th) << endl
         << "AlarmTemperatureLowThresh: "
            << ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature_low_th) << endl
         << "WarnTemperatureHighThresh: " << "N/A" << endl
         << "WarnTemperatureLowThresh: " << "N/A" << endl
         << "AlarmVoltageHighThresh: " << ConvertVoltageToStr(p_module_info->voltage_high_th) << endl
         << "AlarmVoltageLowThresh: " << ConvertVoltageToStr(p_module_info->voltage_low_th) << endl
         << "WarnVoltageHighThresh: N/A" << endl
         << "WarnVoltageLowThresh: N/A";

    IBDIAGNET_RETURN_VOID;
}

PhyCableRecord::ModuleRecord *DiagnosticDataModuleInfo::CreateRecord(const VS_DiagnosticData *p_dd)
{
    IBDIAGNET_ENTER;

    if (!p_dd)
        IBDIAGNET_RETURN(NULL);

    DDModuleInfo module_info;
    DDModuleInfo_unpack(&module_info, (uint8_t *)&(p_dd->data_set));

    IBDIAGNET_RETURN(DiagnosticDataModuleInfo::CreateRecord(&module_info, 0));
}

PhyCableRecord::ModuleRecord *DiagnosticDataModuleInfo::CreateRecord(const struct DDModuleInfo *p_module_info,
                                                                     u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info)
        IBDIAGNET_RETURN(NULL);

    PhyCableRecord::ModuleRecord *p_record = new PhyCableRecord::ModuleRecord();
    DiagnosticDataModuleInfo::CopyTo(*p_module_info, *p_record);
    p_record->module_info_ext = module_info_ext;

    IBDIAGNET_RETURN(p_record);
}

#ifdef COPY
    #error "The COPY macros has been already defined"
#else
    #define COPY(X) to.X = from.X
#endif

#ifdef COPY_ARRAY
    #error "The COPY_ARRAY macros has been already defined"
#else
    #define COPY_ARRAY(X) memcpy(to.X, from.X, sizeof(to.X))
#endif

void DiagnosticDataModuleInfo::CopyTo(const DDModuleInfo &from, PhyCableRecord::ModuleRecord &to)
{
    IBDIAGNET_ENTER;

    COPY(ethernet_compliance_code);
    COPY(ext_ethernet_compliance_code);
    COPY(cable_breakout);
    COPY(cable_technology);
    COPY(cable_power_class);
    COPY(cable_identifier);
    COPY(cable_length);
    COPY(cable_vendor);
    COPY(cable_type);
    COPY(cable_tx_equalization);
    COPY(cable_rx_emphasis);
    COPY(cable_rx_amp);
    COPY(max_power);
    COPY(cable_attenuation_5g);
    COPY(cable_attenuation_7g);
    COPY(cable_attenuation_12g);
    COPY(cable_attenuation_25g);
    COPY(tx_cdr_state);
    COPY(rx_cdr_state);
    COPY(tx_cdr_cap);
    COPY(rx_cdr_cap);
    COPY(cable_rx_post_emphasis);
    COPY_ARRAY(vendor_name);
    COPY_ARRAY(vendor_pn);
    COPY_ARRAY(vendor_rev);
    COPY(fw_version);
    COPY_ARRAY(vendor_sn);
    COPY(voltage);
    COPY(temperature);
    COPY(rx_power_lane1);
    COPY(rx_power_lane0);
    COPY(rx_power_lane3);
    COPY(rx_power_lane2);
    COPY(rx_power_lane5);
    COPY(rx_power_lane4);
    COPY(rx_power_lane7);
    COPY(rx_power_lane6);
    COPY(tx_power_lane1);
    COPY(tx_power_lane0);
    COPY(tx_power_lane3);
    COPY(tx_power_lane2);
    COPY(tx_power_lane5);
    COPY(tx_power_lane4);
    COPY(tx_power_lane7);
    COPY(tx_power_lane6);
    COPY(tx_bias_lane1);
    COPY(tx_bias_lane0);
    COPY(tx_bias_lane3);
    COPY(tx_bias_lane2);
    COPY(tx_bias_lane5);
    COPY(tx_bias_lane4);
    COPY(tx_bias_lane7);
    COPY(tx_bias_lane6);
    COPY(temperature_low_th);
    COPY(temperature_high_th);
    COPY(voltage_low_th);
    COPY(voltage_high_th);
    COPY(rx_power_low_th);
    COPY(rx_power_high_th);
    COPY(tx_power_low_th);
    COPY(tx_power_high_th);
    COPY(tx_bias_low_th);
    COPY(tx_bias_high_th);
    COPY(wavelength);
    COPY(smf_length);
    COPY(did_cap);
    COPY(rx_power_type);
    COPY(module_st);
    COPY(ib_compliance_code);
    COPY(active_set_media_compliance_code);
    COPY(active_set_host_compliance_code);
    COPY(ib_width);
    COPY(monitor_cap_mask);
    COPY(nbr_100);
    COPY(nbr_250);
    COPY(dp_st_lane7);
    COPY(dp_st_lane6);
    COPY(dp_st_lane5);
    COPY(dp_st_lane4);
    COPY(dp_st_lane3);
    COPY(dp_st_lane2);
    COPY(dp_st_lane1);
    COPY(dp_st_lane0);
    COPY(length_om5);
    COPY(length_om4);
    COPY(length_om3);
    COPY(length_om2);
    COPY(memory_map_rev);
    COPY(wavelength_tolerance);
    COPY(length_om1);
    COPY(memory_map_compliance);
    COPY(date_code);
    COPY(vendor_oui);
    COPY(connector_type);
    COPY(rx_output_valid);
    COPY(rx_input_valid);
    COPY(tx_input_freq_sync);
    COPY(error_code);

    IBDIAGNET_RETURN_VOID;
}

#undef COPY
#undef COPY_ARRAY

// PDDR Link Down Info
DiagnosticDataLinkDownInfo::DiagnosticDataLinkDownInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LINK_DOWN_INFO_PAGE,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_VERSION,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_NAME,
                       NOT_SUPPORT_DD_LINK_DOWN_INFO,
                       DD_PHY_TYPE,
                       SECTION_LINK_DOWN_INFO,
                       false)
{
}

void DiagnosticDataLinkDownInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DDLinkDownInfo link_down_info;
    DDLinkDownInfo_unpack(&link_down_info, (uint8_t *)&(dd.data_set));

    sstream << +link_down_info.down_blame << ','
            << +link_down_info.local_reason_opcode << ','
            << +link_down_info.remote_reason_opcode << ','
            << +link_down_info.e2e_reason_opcode << ','
            << +link_down_info.last_ber_mantissa << ','
            << +link_down_info.last_ber_exp << ','
            << +link_down_info.ber_mantissa << ','
            << +link_down_info.ber_exp << ','
            << +link_down_info.min_ber_mantissa << ','
            << +link_down_info.min_ber_exp << ','
            << +link_down_info.max_ber_mantissa << ','
            << +link_down_info.max_ber_exp << ','
            << link_down_info.num_of_ber_alarams;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Link Down Info
DiagnosticDataLatchedFlagInfo::DiagnosticDataLatchedFlagInfo(bool enable_disconnected_ports):
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_PAGE,
                       DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_VERSION,
                       DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NAME,
                       NOT_SUPPORT_DD_LATCHED_FLAG_INFO,
                       DD_PHY_TYPE,
                       SECTION_LATCHED_FLAG_INFO,
                       false,
                       SUPPORT_SW,
                       enable_disconnected_ports)
{
}

void DiagnosticDataLatchedFlagInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct DDLatchedFlagInfo latched_flag_info;
    DDLatchedFlagInfo_unpack(&latched_flag_info, (uint8_t *)&(dd.data_set));

    DumpLatchedFlagInfoRawData(sstream, latched_flag_info);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoRawData(stringstream &sstream, const struct DDLatchedFlagInfo &latched_flag_info)
{
    IBDIAGNET_ENTER;

    sstream << +latched_flag_info.dp_fw_fault << ','
            << +latched_flag_info.mod_fw_fault << ','
            << +latched_flag_info.vcc_flags << ','
            << +latched_flag_info.temp_flags << ','
            << +latched_flag_info.tx_ad_eq_fault << ','
            << +latched_flag_info.tx_cdr_lol << ','
            << +latched_flag_info.tx_los << ','
            << +latched_flag_info.tx_fault << ','
            << +latched_flag_info.tx_power_lo_war << ','
            << +latched_flag_info.tx_power_hi_war << ','
            << +latched_flag_info.tx_power_lo_al << ','
            << +latched_flag_info.tx_power_hi_al << ','
            << +latched_flag_info.tx_bias_lo_war << ','
            << +latched_flag_info.tx_bias_hi_war << ','
            << +latched_flag_info.tx_bias_lo_al << ','
            << +latched_flag_info.tx_bias_hi_al << ','
            << +latched_flag_info.rx_cdr_lol << ','
            << +latched_flag_info.rx_los << ','
            << +latched_flag_info.rx_power_lo_war << ','
            << +latched_flag_info.rx_power_hi_war << ','
            << +latched_flag_info.rx_power_lo_al << ','
            << +latched_flag_info.rx_power_hi_al << ','
            << +latched_flag_info.rx_output_valid_change << ','
            << +latched_flag_info.rx_input_valid_change;

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoHeader(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "RX1LatchedLossIndicator,RX2LatchedLossIndicator,RX3LatchedLossIndicator,"
            << "RX4LatchedLossIndicator,TX1LatchedLossIndicator,TX2LatchedLossIndicator,"
            << "TX3LatchedLossIndicator,TX4LatchedLossIndicator,"
            << "TX1AdaptiveEqualizationFaultIndicator,"
            << "TX2AdaptiveEqualizationFaultIndicator,TX3AdaptiveEqualizationFaultIndicator,"
            << "TX4AdaptiveEqualizationFaultIndicator,RX1CDRLOL,RX2CDRLOL,RX3CDRLOL,RX4CDRLOL,"
            << "TX1CDRLOL,TX2CDRLOL,TX3CDRLOL,TX4CDRLOL,"
            << "HighTemperatureAlarm,LowTemperatureAlarm,"
            << "HighTemperatureWarning,LowTemperatureWarning,"
            << "HighSupplyVoltageAlarm,LowSupplyVoltageAlarm,HighSupplyVoltageWarning,"
            << "LowSupplyVoltageWarning,"
            << "HighRX1PowerAlarm,LowRX1PowerAlarm,HighRX1PowerWarning,LowRX1PowerWarning,"
            << "HighRX2PowerAlarm,LowRX2PowerAlarm,HighRX2PowerWarning,LowRX2PowerWarning,"
            << "HighRX3PowerAlarm,LowRX3PowerAlarm,HighRX3PowerWarning,LowRX3PowerWarning,"
            << "HighRX4PowerAlarm,LowRX4PowerAlarm,HighRX4PowerWarning,LowRX4PowerWarning,"
            << "HighTX1BiasAlarm,LowTX1BiasAlarm,HighTX1BiasWarning,LowTX1BiasWarning,"
            << "HighTX2BiasAlarm,LowTX2BiasAlarm,HighTX2BiasWarning,LowTX2BiasWarning,"
            << "HighTX3BiasAlarm,LowTX3BiasAlarm,HighTX3BiasWarning,LowTX3BiasWarning,"
            << "HighTX4BiasAlarm,LowTX4BiasAlarm,HighTX4BiasWarning,LowTX4BiasWarning,"
            << "HighTX1PowerAlarm,LowTX1PowerAlarm,HighTX1PowerWarning,LowTX1PowerWarning,"
            << "HighTX2PowerAlarm,LowTX2PowerAlarm,HighTX2PowerWarning,LowTX2PowerWarning,"
            << "HighTX3PowerAlarm,LowTX3PowerAlarm,HighTX3PowerWarning,LowTX3PowerWarning,"
            << "HighTX4PowerAlarm,LowTX4PowerAlarm,HighTX4PowerWarning,LowTX4PowerWarning";

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(stringstream &sstream,
                                                            VS_DiagnosticData *dd)
{
    IBDIAGNET_ENTER;

    struct DDLatchedFlagInfo latched_flag_info;
    struct DDLatchedFlagInfo *p_latched_flag_info = NULL;

    if (dd)
        DDLatchedFlagInfo_unpack(p_latched_flag_info = &latched_flag_info, (uint8_t *)&(dd->data_set));

    DumpLatchedFlagInfoData(sstream, p_latched_flag_info);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(stringstream &sstream,
                                                            const struct DDLatchedFlagInfo *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info) {
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A";
        IBDIAGNET_RETURN_VOID;
    }


    sstream << +(p_latched_flag_info->rx_los & 0x1)                                              << ','
            << +((p_latched_flag_info->rx_los >> 1) & 0x1)                                       << ','
            << +((p_latched_flag_info->rx_los >> 2) & 0x1)                                       << ','
            << +((p_latched_flag_info->rx_los >> 3) & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_los & 0x1)                                              << ','
            << +((p_latched_flag_info->tx_los >> 1) & 0x1)                                       << ','
            << +((p_latched_flag_info->tx_los >> 2) & 0x1)                                       << ','
            << +((p_latched_flag_info->tx_los >> 3) & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_ad_eq_fault & 0x1)                                      << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 3) & 0x1)                               << ','
            << +(p_latched_flag_info->rx_cdr_lol & 0x1)                                          << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->tx_cdr_lol & 0x1)                                          << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->temp_flags & 0x1)                                          << ','
            << +((p_latched_flag_info->temp_flags >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->temp_flags >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->temp_flags >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->vcc_flags & 0x1)                                           << ','
            << +((p_latched_flag_info->vcc_flags >> 1) & 0x1)                                    << ','
            << +((p_latched_flag_info->vcc_flags >> 2) & 0x1)                                    << ','
            << +((p_latched_flag_info->vcc_flags >> 3) & 0x1)                                    << ','
            << +(p_latched_flag_info->rx_power_hi_al & 0x1)                                      << ','
            << +(p_latched_flag_info->rx_power_lo_al & 0x1)                                      << ','
            << +(p_latched_flag_info->rx_power_hi_war & 0x1)                                     << ','
            << +(p_latched_flag_info->rx_power_lo_war & 0x1)                                     << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 3) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 3) & 0x1)                              << ','
            << +(p_latched_flag_info->tx_bias_hi_al & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_bias_lo_al & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_bias_hi_war & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_bias_lo_war & 0x1)                                      << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 1) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 1) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 2) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 2) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 3) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 3) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 3) & 0x1)                               << ','
            << +(p_latched_flag_info->tx_power_hi_al & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_power_lo_al & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_power_hi_war & 0x1)                                     << ','
            << +(p_latched_flag_info->tx_power_lo_war & 0x1)                                     << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 3) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 3) & 0x1);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(ofstream &sout,
                                                            VS_DiagnosticData *dd)
{
    IBDIAGNET_ENTER;

    struct DDLatchedFlagInfo latched_flag_info;
    struct DDLatchedFlagInfo *p_latched_flag_info = NULL;

    if (dd)
        DDLatchedFlagInfo_unpack(p_latched_flag_info = &latched_flag_info, (uint8_t *)&(dd->data_set));

    DumpLatchedFlagInfoData(sout, p_latched_flag_info);

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(ofstream &sout,
                                                            const struct DDLatchedFlagInfo *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info) {
         sout << "CDR Latched RX Loss Indicator: N/A" << endl
              << "CDR Latched TX Loss Indicator: N/A" << endl
              << "Latched Adaptive Equalization Fault: N/A" << endl
              << "Latched  RX LOL Indicator: N/A" << endl
              << "Latched  TX LOL Indicator: N/A" << endl
              << "Latched Temperature Alarm and Warning: N/A" << endl
              << "Latched Voltage Alarm and Warning: N/A" << endl
              << "RX Power High Alarm: N/A" << endl
              << "RX Power Low Alarm: N/A" << endl
              << "RX Power High Warning: N/A" << endl
              << "RX Power Low Warning: N/A" << endl
              << "TX Bias High Alarm: N/A" << endl
              << "TX Bias Low Alarm: N/A" << endl
              << "TX Bias High Warning: N/A" << endl
              << "TX Bias Low Warning: N/A" << endl
              << "TX Power High Alarm: N/A" << endl
              << "TX Power Low Alarm: N/A" << endl
              << "TX Power High Warning: N/A" << endl
              << "TX Power Low Warning: N/A";

         IBDIAGNET_RETURN_VOID;
    }

    sout << "CDR Latched RX Loss Indicator: " << +(p_latched_flag_info->rx_los & 0xf) << endl
         << "CDR Latched TX Loss Indicator: " << +(p_latched_flag_info->tx_los & 0xf) << endl
         << "Latched Adaptive Equalization Fault: "
             << +(p_latched_flag_info->tx_ad_eq_fault & 0xf) << endl
         << "Latched  RX LOL Indicator: " << +(p_latched_flag_info->rx_cdr_lol & 0xf) << endl
         << "Latched  TX LOL Indicator: " << +(p_latched_flag_info->tx_cdr_lol & 0xf) << endl
         << "Latched Temperature Alarm and Warning: "
             << +(p_latched_flag_info->temp_flags & 0xf) << endl
         << "Latched Voltage Alarm and Warning: " << +(p_latched_flag_info->vcc_flags & 0xf) << endl
         << "RX Power High Alarm: " << +(p_latched_flag_info->rx_power_hi_al & 0xf) << endl
         << "RX Power Low Alarm: " << +(p_latched_flag_info->rx_power_lo_al & 0xf) << endl
         << "RX Power High Warning: " << +(p_latched_flag_info->rx_power_hi_war & 0xf) << endl
         << "RX Power Low Warning: " << +(p_latched_flag_info->rx_power_lo_war & 0xf) << endl
         << "TX Bias High Alarm: " << +(p_latched_flag_info->tx_bias_hi_al & 0xf) << endl
         << "TX Bias Low Alarm: " << +(p_latched_flag_info->tx_bias_lo_al & 0xf) << endl
         << "TX Bias High Warning: " << +(p_latched_flag_info->tx_bias_hi_war & 0xf) << endl
         << "TX Bias Low Warning: " << +(p_latched_flag_info->tx_bias_lo_war & 0xf) << endl
         << "TX Power High Alarm: " << +(p_latched_flag_info->tx_power_hi_al & 0xf) << endl
         << "TX Power Low Alarm: " << +(p_latched_flag_info->tx_power_lo_al & 0xf) << endl
         << "TX Power High Warning: " << +(p_latched_flag_info->tx_power_hi_war & 0xf) << endl
         << "TX Power Low Warning: " << +(p_latched_flag_info->tx_power_lo_war & 0xf);

    IBDIAGNET_RETURN_VOID;
}

PhyCableRecord::LatchedRecord
                    *DiagnosticDataLatchedFlagInfo::CreateRecord(const VS_DiagnosticData *p_dd)
{
    IBDIAGNET_ENTER;

    if (!p_dd)
        IBDIAGNET_RETURN(NULL);

    DDLatchedFlagInfo latched_flag_info;
    DDLatchedFlagInfo_unpack(&latched_flag_info, (uint8_t *)&(p_dd->data_set));

    IBDIAGNET_RETURN(DiagnosticDataLatchedFlagInfo::CreateRecord(&latched_flag_info));
}

PhyCableRecord::LatchedRecord
    *DiagnosticDataLatchedFlagInfo::CreateRecord(const struct DDLatchedFlagInfo *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info)
        IBDIAGNET_RETURN(NULL);

    PhyCableRecord::LatchedRecord *p_latched = new PhyCableRecord::LatchedRecord();
    DiagnosticDataLatchedFlagInfo::CopyTo(*p_latched_flag_info, *p_latched);

    IBDIAGNET_RETURN(p_latched);
}

#ifdef COPY
    #error "The COPY macros has been already defined"
#else
    #define COPY(X) to.X = from.X
#endif

void DiagnosticDataLatchedFlagInfo::CopyTo(const DDLatchedFlagInfo &from , PhyCableRecord::LatchedRecord &to)
{
    IBDIAGNET_ENTER;

    COPY(dp_fw_fault);
    COPY(mod_fw_fault);
    COPY(vcc_flags);
    COPY(temp_flags);
    COPY(tx_ad_eq_fault);
    COPY(tx_cdr_lol);
    COPY(tx_los);
    COPY(tx_fault);
    COPY(tx_power_lo_war);
    COPY(tx_power_hi_war);
    COPY(tx_power_lo_al);
    COPY(tx_power_hi_al);
    COPY(tx_bias_lo_war);
    COPY(tx_bias_hi_war);
    COPY(tx_bias_lo_al);
    COPY(tx_bias_hi_al);
    COPY(rx_cdr_lol);
    COPY(rx_los);
    COPY(rx_power_lo_war);
    COPY(rx_power_hi_war);
    COPY(rx_power_lo_al);
    COPY(rx_power_hi_al);
    COPY(rx_output_valid_change);
    COPY(rx_input_valid_change);

    IBDIAGNET_RETURN_VOID;
}

#undef COPY

// PDDR Link Down Info
DiagnosticDataLinkUpInfo::DiagnosticDataLinkUpInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LINK_UP_INFO_PAGE,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_VERSION,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_NAME,
                       NOT_SUPPORT_DD_LINK_UP_INFO,
                       DD_PHY_TYPE,
                       SECTION_LINK_UP_INFO,
                       false)
{
}

void DiagnosticDataLinkUpInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DDLinkUpInfo link_up_info;
    DDLinkUpInfo_unpack(&link_up_info, (uint8_t *)&(dd.data_set));

    sstream << +link_up_info.up_reason_mng << ','
            << +link_up_info.up_reason_drv << ','
            << +link_up_info.up_reason_pwr << ','
            << +link_up_info.time_to_link_up << ','
            << +link_up_info.fast_link_up_status << ','
            << +link_up_info.time_to_link_up_phy_up_to_active << ','
            << +link_up_info.time_to_link_up_sd_to_phy_up << ','
            << +link_up_info.time_to_link_up_disable_to_sd << ','
            << +link_up_info.time_to_link_up_disable_to_pd;

    IBDIAGNET_RETURN_VOID;
}


// PPCNT pages
// PPCNT RS FEC Histograms
DiagnosticDataRSHistograms::DiagnosticDataRSHistograms():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_PAGE,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_VERSION,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NAME,
                       NOT_SUPPORT_DD_RS_FEC_HISTOGRAM,
                       DD_PHY_TYPE,
                       SECTION_RS_FEC_HISTOGRAM,
                       false)
{
}

void DiagnosticDataRSHistograms::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DD_RS_Histograms rs_histograms;
    DD_RS_Histograms_unpack(&rs_histograms, (uint8_t *)&(dd.data_set));

    sstream << rs_histograms.hist[0];
    for (int i = 1; i < DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS; ++i) {
        sstream << ',' << rs_histograms.hist[i];
    }

    IBDIAGNET_RETURN_VOID;
}

// PPCNT PLR Counters
DiagnosticDataPLRCounters::DiagnosticDataPLRCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PLR_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_NAME,
                       NOT_SUPPORT_DD_PLR_COUNTERS,
                       DD_PHY_TYPE,
                       SECTION_PLR_COUNTERS,
                       false)
{
}

void DiagnosticDataPLRCounters::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                   IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DD_PLR_Counters plr_counters;
    DD_PLR_Counters_unpack(&plr_counters, (uint8_t *)&(dd.data_set));

    sstream << plr_counters.plr_rcv_codes << ','
            << plr_counters.plr_rcv_code_err << ','
            << plr_counters.plr_rcv_uncorrectable_code << ','
            << plr_counters.plr_xmit_codes << ','
            << plr_counters.plr_xmit_retry_codes << ','
            << plr_counters.plr_xmit_retry_events << ','
            << plr_counters.plr_sync_events << ','
            << plr_counters.hi_retransmission_rate << ",";

    if (this->m_p_ibdiag->GetCapabilityModulePtr()->
                          IsSupportedGMPCapability(p_node,
                                                   EnGMPCAPIsPLRMaxRetransmissionRateSupported))
        sstream << DEC(plr_counters.plr_xmit_retry_codes_within_t_sec_max);
    else
        sstream << "N/A";

    IBDIAGNET_RETURN_VOID;
}


// PPCNT Phy Statistics
DiagnosticDataPhyStatistics::DiagnosticDataPhyStatistics():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_VERSION,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NAME,
                       NOT_SUPPORT_DD_PHY_LAYER_STATISTICS,
                       DD_PHY_TYPE,
                       SECTION_PHY_LAYER_STATISTICS,
                       false,
                       SUPPORT_SW_CA)
{
}

void DiagnosticDataPhyStatistics::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                     IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DD_PhyStatisticalCounters phy_statistics;
    DD_PhyStatisticalCounters_unpack(&phy_statistics, (uint8_t *)&(dd.data_set));

    sstream << phy_statistics.time_since_last_clear << ','
            << phy_statistics.phy_received_bits << ','
            << phy_statistics.phy_symbol_errors << ','
            << phy_statistics.phy_corrected_bits << ','
            << phy_statistics.phy_raw_errors_lane0<< ','
            << phy_statistics.phy_raw_errors_lane1 << ','
            << phy_statistics.phy_raw_errors_lane2 << ','
            << phy_statistics.phy_raw_errors_lane3 << ','
            << phy_statistics.phy_raw_errors_lane4 << ','
            << phy_statistics.phy_raw_errors_lane5 << ','
            << phy_statistics.phy_raw_errors_lane6 << ','
            << phy_statistics.phy_raw_errors_lane7 << ','
            << +phy_statistics.raw_ber_coef << ','
            << +phy_statistics.raw_ber_magnitude << ','
            << +phy_statistics.effective_ber_coef << ','
            << +phy_statistics.effective_ber_magnitude << ",";

    if (this->m_p_ibdiag->GetCapabilityModulePtr()->
                          IsSupportedGMPCapability(p_node, EnGMPCAPIsEffectiveCounterSupported))

        sstream << DEC(phy_statistics.symbol_ber_coef)        << ","
                << DEC(phy_statistics.symbol_ber_magnitude)   << ","
                << DEC(phy_statistics.phy_effective_errors)   << ",";
    else
        sstream << "N/A,N/A,N/A,";

    if (this->m_p_ibdiag->GetCapabilityModulePtr()->
            IsSupportedGMPCapability(p_node, EnGMPCAPIsRawBerPerLaneSupported))

        sstream << +phy_statistics.raw_ber_coef_lane0        << ','
                << +phy_statistics.raw_ber_magnitude_lane0   << ','
                << +phy_statistics.raw_ber_coef_lane1        << ','
                << +phy_statistics.raw_ber_magnitude_lane1   << ','
                << +phy_statistics.raw_ber_coef_lane2        << ','
                << +phy_statistics.raw_ber_magnitude_lane2   << ','
                << +phy_statistics.raw_ber_coef_lane3        << ','
                << +phy_statistics.raw_ber_magnitude_lane3   << ','
                << +phy_statistics.raw_ber_coef_lane4        << ','
                << +phy_statistics.raw_ber_magnitude_lane4   << ','
                << +phy_statistics.raw_ber_coef_lane5        << ','
                << +phy_statistics.raw_ber_magnitude_lane5   << ','
                << +phy_statistics.raw_ber_coef_lane6        << ','
                << +phy_statistics.raw_ber_magnitude_lane6   << ','
                << +phy_statistics.raw_ber_coef_lane7        << ','
                << +phy_statistics.raw_ber_magnitude_lane7   ;

    else
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A";

    IBDIAGNET_RETURN_VOID;
}

// PPCNT 0x26 InfiniBand General Counters
DiagnosticDataInfiniBandGeneralCounters::DiagnosticDataInfiniBandGeneralCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NAME,
                       NOT_SUPPORT_DD_INFINIBAND_GENERAL_COUNTERS,
                       DD_PHY_TYPE,
                       SECTION_INFINIBAND_GENERAL_COUNTERS,
                       false,
                       SUPPORT_SW_CA)
{
}

void DiagnosticDataInfiniBandGeneralCounters::DumpDiagnosticData(stringstream &sstream,
                                                                 VS_DiagnosticData &dd,
                                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct DD_InfiniBandGeneralCounter data;
    DD_InfiniBandGeneralCounter_unpack(&data, (uint8_t *)&(dd.data_set));

    sstream << HILOW_TO_UINT64(data.port_rcv_data_qword) << ','
            << HILOW_TO_UINT64(data.port_xmit_data_qword) << ','
            << HILOW_TO_UINT64(data.rq_general_error) << ','
            << HILOW_TO_UINT64(data.port_xmit_ebp) << ','
            << HILOW_TO_UINT64(data.port_sw_hoq_lifetime_limit_discards) << ','
            << HILOW_TO_UINT64(data.dqs2llu_xmit_wait_arb_global) << ','
            << HILOW_TO_UINT64(data.grxb_fecn_mark) << ','
            << HILOW_TO_UINT64(data.general_transmit_discard_external_contain);

    IBDIAGNET_RETURN_VOID;
}
