/*
 * Copyright (c) 2004-2010 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
 * Fabric Topology Matcher Test Program
 *
 * Reads a Topology file and OpenSM subnet file, try to match and report errors.
 */


#include <getopt.h>

#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SimulateA15.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/CredLoops.h>
#include <infiniband/ibdm/TopoMatch.h>
#include <infiniband/ibutils/release_containers.h>

#define IBNL_ENV_VAR_NAME "IBDM_IBNL_PATH"

static char TopoMatchTestUsage[] =
"Usage: ibtopodiff [-v][-h][-V][-m map-file][-c csv-file]"
" -t <topo file> -d <subnet file> [-D max-dist] {-e | -s <start node name> -p <start port num> -g <start port guid>}" ;

void
show_usage() {
  cout << TopoMatchTestUsage << endl;
}

void
show_help() {
  cout << "\n"
       << " Fabric Topology Matcher\n"
       << "-------------------------------\n"
       << "\n"
       << TopoMatchTestUsage << "\n"
       << "\n"
       << "Description:\n"
       << "This tool compares a topology file and a discovered listing of "
       << "lst or ibnetdiscover files and reports missmatches.\n"
       << "Three different algorithms provided:\n"
       << "Using the -e option is more suitible for MANY missmatches\n"
       << "it applies less heuristics and provide details about the match.\n"
       << "Using the -D option is aimed to verify the dicovered topology"
       << " around the start node and up to given distance.\n"
       << "Providing the -s, -p and -g starts a detailed heuristics that\n"
       << "should be used when only small number of changes are expected.\n"
       << "\n"
       << "Arguments:\n"
       << "  -t|--topology <file> = Topology file (topo or lst or ibnetdiscover).\n"
       << "       topo file format is defined in the IBDM user manual.\n"
       << "  -d|--discovered <file> = lst file ('.lst') or ibnetdiscover file ('.ibnetdiscover' or '.ibnd' or '.net').\n"
       << "  -e|--edge = start processing from the edges using strict match.\n"
       << "  -D|--distance = only report missconnected nodes/ports up to requested distance from the starting node.\n"
       << "  -n|--smart-name = trying search switch node by description if node wasn't found by name (experimental).\n"
       << "  -s|--start-node <name> = The name of the start node [H-1/U1].\n"
       << "  -p|--port-num <num> = The number of the start port [1].\n"
       << "  -g|--port-guid <guid> = The guid of the start port [none].\n"
       << "\n"
       << "Options:\n"
       << "  -c|--csv out-file = optionally output into the given CSV file.\n"
       << "  -m|--map map-file = mapping topo node to discover node name/GUID\n"
       << "  -v|--verbose = verbsoe mode\n"
       << "  -h|--help = provide this help message\n"
       << "  -V|--version = Prints the version of the tool\n"
       << "\n"
       << "\n"
       << "Author: Eitan Zahavi, Mellanox Technologies LTD.\n"
       << endl;
}


int main (int argc, char **argv) {
  /*
   * Parseing of Command Line
   */
  enum RunMode {GraphMatch, EdgeBased, CenterNode};

  const char * startSystemName = "H-1";
  const char * topoFileName = "ibdm.topo";
  const char * subnetFileName = "subnet.lst";
  string       mapFileName;
  string       csvFileName;
  phys_port_t startPortNum = 1;
  uint64_t startPortGuid = 0;
  RunMode runMode = GraphMatch;
  char *p_diagnostics;
  int next_option;
  int rc = 0;
  bool smartSWsName = false;
  unsigned int centerModeDepth = 1;
  const char * const short_option = "vhs:p:g:d:t:VeD:nm:c:i:";
  /*
	 In the array below, the 2nd parameter specified the number
	 of arguments as follows:
	 0: no arguments
	 1: argument
	 2: optional
  */
  const option long_option[] =
         {
                {       "verbose",       0,     NULL,   'v'},
                {       "help",          0,     NULL,   'h'},
                {       "csv",           1,     NULL,   'c'},
                {       "start-system",  1,     NULL,   's'},
                {       "port-num",      1,     NULL,   'p'},
                {       "port-guid",     1,     NULL,   'g'},
                {       "discovered",    1,     NULL,   'd'},
                {       "topology",      1,     NULL,   't'},
                {       "ibnl_dir",      1,     NULL,   'i'},
                {       "version",       0,     NULL,   'V'},
                {       "edge",          0,     NULL,   'e'},
                {       "dist",          1,     NULL,   'D'},
                {       "smart-names",   0,     NULL,   'n'},
                {       "map",           1,     NULL,   'm'},
                {       NULL,            0,     NULL,    0 }    /* Required at the end of the array */
         };

  printf("-------------------------------------------------\n");
  do
  {
	 next_option = getopt_long(argc, argv, short_option, long_option, NULL);
	 switch(next_option)
	 {
	 case 'v':
		/*
		  Verbose Mode
		*/
		FabricUtilsVerboseLevel |= FABU_LOG_VERBOSE;
		printf(" Verbose Mode\n");
		break;

	 case 's':
		/*
		  Specifies Start System
		*/
		startSystemName = optarg;
		break;

	 case 'd':
		/*
		  Specifies Discovered subnet.lst file name
		*/
		subnetFileName = optarg;
		break;

	 case 'p':
		/*
		  Specifies Start Port Num
		*/
		startPortNum = (phys_port_t)atoi(optarg);
		break;

	 case 'm':
		/*
		  Specifies Mapping File
		*/
	        mapFileName = string(optarg);
		break;

	 case 'c':
		/*
		  Specifies CSV out File
		*/
	        csvFileName = string(optarg);
		break;

	 case 'n':
		/*
		  Specifies smart names find for switches
		*/
		smartSWsName = true;
		break;

	 case 'g':
		/*
		  Specifies Start Port Guid
		*/
		startPortGuid = strtoull(optarg, NULL, 16);
		break;

	 case 't':
		/*
		  Specifies Subnet Cabling file
		*/
		topoFileName = optarg;
		break;

	 case 'e':
		/*
		  Edge Based Algorithm
		*/
		runMode = EdgeBased;
		break;

	 case 'D':
		/*
		  Center Mode Distance Based Algorithm
		*/
		runMode = CenterNode;
		centerModeDepth = atoi(optarg);
		break;

   case 'i':
		if (setenv(IBNL_ENV_VAR_NAME, optarg, 1)) {
                    printf("Error with set ibnl direcory path\n");
                    return 1;
                }
		break;

	 case 'V':
		printf("ibtopodiff version: %s\n", get_ibdm_version());
		return 0;

	 case 'h':
		show_help();
		return 0;
		break;

	 case -1:
		break; /* done with option */
	 default: /* something wrong */
		show_usage();
		exit(1);
	 }
  }
  while(next_option != -1);

  if (runMode == GraphMatch) {
    if (!startPortGuid) {
	    printf("-E- Missing -g/--port-guid mandatory argument for graph based match.\n");
	    show_usage();
	    exit(1);
    }
  }

  printf(" Fabric Topology Match:\n");
  printf(" Topology File .. %s\n", topoFileName);
  printf(" Subnet File .... %s\n", subnetFileName);
  if (runMode != EdgeBased) {
    if (runMode == GraphMatch) {
      printf(" Running graph based match heuristics.\n");
    } else {
      printf(" Running node centric match heuristics.\n");
    }
    printf(" Start Node ..... %s\n", startSystemName);
    printf(" Start Port ..... %u\n", startPortNum);
    if (startPortGuid)
      printf(" Start Guid ..... 0x%016" PRIx64 "\n", startPortGuid);
  } else {
    printf(" Running edge mode match heuristics.\n");
  }
  printf("-------------------------------------------------\n");

  IBFabric *p_dFabric, *p_sFabric;
  IBFabric *p_mFabric;
  string subnetFileName_str(subnetFileName);
  string ext;
  IBDiag* p_sIbdiag = new IBDiag();
  IBDiag* p_dIbdiag = new IBDiag();
  list_p_fabric_general_err aports_errors;

  p_sFabric = p_sIbdiag->GetDiscoverFabricPtr();
  p_dFabric = p_dIbdiag->GetDiscoverFabricPtr();
  p_mFabric = new IBFabric();

  if (p_sFabric->parseTopology(topoFileName)) {
    cout << "-E- Fail to parse specified topology file:" << topoFileName << endl;
    exit(1);
  }

  for (auto& name_sys_pair : p_sFabric->SystemByName) {
    if (SimulateA15::SimulateSystemHeirarchyInfo(name_sys_pair.second)) {
      cout << "-W- Failed to create hierarchy info for system "
           << name_sys_pair.second->name << endl;
      continue;
    }
  }

  if (p_sFabric->AllocateFabricNodeGuids())
    cout << "-W- Failed to create node and system GUIDs for .topo fabric" << endl;

  p_sIbdiag->CollectAPortsData(aports_errors);
  if (!aports_errors.empty()) {
    cout << "-W- Failed to create APorts from .topo file. Errors: " << endl;
    for (FabricErrGeneral* err : aports_errors) {
      if (!err)
        continue;
      cout << err->GetErrorLine() << endl;
      delete err;
    }
    aports_errors.clear();
  }

  //s_fabric.dump(cout);

  // get the extension of the file name
  size_t pos = subnetFileName_str.rfind(".");
  if (pos == string::npos) {
    cout << "-E- File extension not found for " << subnetFileName_str << endl;
    exit(1);
  }
  ext = subnetFileName_str.substr(pos + 1);

  if (ext == "db_csv") {
    // attempt to build p_dFabric from db_csv
    try{
      rc = p_dIbdiag->DiscoverFabricFromFile(subnetFileName_str, false);
    } catch(...) {
      rc = 1;
    }
    if (rc) {
      cout << "-E- Failed to create fabric from db_csv file: " << subnetFileName_str << endl;
      exit(1);
    }
    if (p_dIbdiag->CollectAPortsData(aports_errors) || aports_errors.size() > 0) {
     cout << "Failed to create APorts from db_csv file: "  << subnetFileName_str << endl;
     exit(1);
    }
  } else {
    // attempt to build fabric from lst or ibnetdiscover
    if (p_dFabric->parseTopology(subnetFileName, false)) {
      cout << "-E- Fail to parse discovered topology file:" << subnetFileName << endl;
      exit(1);
    }
  }

  int matchStatus;
  if (runMode == GraphMatch) {
    matchStatus =
      TopoMatchFabrics(p_sFabric, p_dFabric, startSystemName, startPortNum,
		       startPortGuid, &p_diagnostics, 0);
  } else if (runMode == EdgeBased) {
    TopoMatchOptions opts(mapFileName,
                          csvFileName,
                          "ibtopodiff.names",
                          true, // matchCAsByName
                          true, // matchSWsByName
                          smartSWsName,
                          centerModeDepth);

    matchStatus =
      TopoMatchFabricsFromEdge(p_sFabric, p_dFabric, opts, &p_diagnostics);
  } else { // runMode == CenterNode)
    TopoMatchOptions opts(mapFileName,
                          csvFileName,
                          "ibtopodiff.names",
                          true, // matchCAsByName
                          true, // matchSWsByName
                          smartSWsName,
                          centerModeDepth);
    matchStatus =
       TopoMatchFabricsFromNode(p_sFabric, p_dFabric, startSystemName, startPortNum,
		       startPortGuid, opts, &p_diagnostics);
  }

  if (p_diagnostics)
  {
    printf("%s\n", p_diagnostics);
    free(p_diagnostics);
  }
  if (matchStatus) {
    printf("-E- Fabric Topologies Differ !\n");
  } else {
    printf("-I- Fabric Topologies Match\n");
  }

  if (runMode == CenterNode)
    exit(matchStatus);

  if (TopoMergeDiscAndSpecFabrics( p_sFabric, p_dFabric, p_mFabric)) {
    printf("-E- Fail to merge fabrics.\n");
    exit(1);
  }

  if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
    p_mFabric->dump(cout);

  delete p_sIbdiag;
  delete p_dIbdiag;
  delete p_mFabric;
  release_container_data(aports_errors);

  exit(0);
}
