/*
 * Copyright (c) 2020-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <errno.h>
#include <sys/stat.h>

#include <infiniband/ibdm/OutputControl.h>

#define OUTPUT_CONTROL_APP_DEFAULT_NAME         "<app-default>"
#define OUTPUT_CONTROL_ALL_NAME                 "all"
#define OUTPUT_CONTROL_DEFAULT_NAME             "default"

const OutputControl::Identity OutputControl::Identity::Null;

OutputControl::Identity::Identity(Flags flags)
    : m_flags(Flags(flags | OutputControl_Flag_Valid))
{
    switch(flags & OutputControl_Flag_Special_Mask)
    {
        case OutputControl_Flag_AppDefault:
            m_type = OUTPUT_CONTROL_APP_DEFAULT_NAME;
            break;

        case OutputControl_Flag_All:
            m_type = OUTPUT_CONTROL_ALL_NAME;
            break;

        case OutputControl_Flag_Default:
            m_type = OUTPUT_CONTROL_DEFAULT_NAME;
            break;

        default:
            m_flags = OutputControl_Flag_None;
            return;
    }

    if(!build_key()) {
        m_flags = OutputControl_Flag_None;
        return;
    }
}

OutputControl::Identity::Identity(const std::string & text, Flags flags)
    : m_flags(OutputControl_Flag_None), m_text(text)
{
    // user flag validation, both flags cannot be used together
    if(flags != OutputControl_Flag_None)
        if((flags & OutputControl_Flag_File_Mask) == OutputControl_Flag_File_Mask)
            return;

    std::string                tkey               = trim(text);
    std::string::size_type     type_index         = tkey.find_last_of('.');
    std::string::size_type     filename_index     = tkey.find_last_of("\\/");

    if(tkey.size() > 3 &&
        (tkey[0] == 'C' || tkey[0] == 'c') &&
        (tkey[1] == 'S' || tkey[1] == 's') &&
        (tkey[2] == 'V' || tkey[2] == 'v') && tkey[3] == ':')
    {
        if(type_index != std::string::npos || filename_index != std::string::npos || flags)
            return;

        m_flags     = OutputControl_Flag_CSV;
        type_index  = 4;
    }
    else
        if(flags == OutputControl_Flag_UserFile) {
            m_flags = Flags(OutputControl_Flag_UserFile |
                            OutputControl_Flag_Valid |
                            OutputControl_Flag_Generic);
            return;
        }
        else
        {
            if(filename_index != std::string::npos)
                if(type_index == std::string::npos || filename_index > type_index) {
                    m_flags = OutputControl_Flag_None;
                    return;
                }

            if(type_index == std::string::npos)
                type_index = 0;
            else
                ++type_index;

            m_flags = Flags(m_flags | OutputControl_Flag_Generic);
        }

    if(tkey.begin() + type_index < tkey.end())
    {
        std::transform(
                tkey.begin() + type_index,
                tkey.end(),
                std::back_inserter(m_type),
                ::tolower);

        if(m_type == OUTPUT_CONTROL_ALL_NAME)
            m_flags = Flags(m_flags | OutputControl_Flag_All);
        else
            if(m_type == OUTPUT_CONTROL_DEFAULT_NAME)
                m_flags = Flags(m_flags | OutputControl_Flag_Default);

        m_flags = Flags(m_flags | OutputControl_Flag_Valid);

        if(!build_key()) {
            m_flags = OutputControl_Flag_None;
            return;
        }
    }
    else
        m_flags = OutputControl_Flag_None;

    // update flags
    m_flags = Flags(m_flags | flags);
}

bool OutputControl::Identity::build_key()
{
    switch(m_flags & OutputControl_Flag_Type_Mask)
    {
        case OutputControl_Flag_Generic:
            m_key = m_type;
            return true;

        case OutputControl_Flag_CSV:
            m_key = "csv:" + m_type;
            return true;

        default:
            return false;
    }
}

std::ostream & OutputControl::Identity::output(std::ostream & stream, const std::string & prefix) const
{
    stream << prefix << "Identity: flags=0x"
            << std::hex << std::setw(8) << std::setfill('0') << m_flags
            << std::dec << std::setfill(' ')
            << " [" << (is_valid() ? "Valid" : "Invalid");

    if(is_csv())
        stream << "|CSV";

    if(is_generic())
        stream << "|Generic";

    if(is_special())
        stream << "|Special";

    if(is_all())
        stream << "|All";

    if(is_default())
        stream << "|Default";

    if(is_app_default())
        stream << "|APP";

    stream << "] type='" << m_type << "' key='" << m_key << "' text='" << m_text<< "'";

    return stream;
}

std::string OutputControl::Identity::to_string() const
{
    std::stringstream ss;
    output(ss, "");
    return ss.str();
}

OutputControl::OutputControl()
    : m_enabled(m_aliases, "Enable output", Flags(OutputControl_Flag_CSV | OutputControl_Flag_Generic)),
      m_pathes(m_aliases, "Path per file", Flags(OutputControl_Flag_CSV | OutputControl_Flag_Generic)),
      m_csv_to_file(m_aliases, "Extract CSV section to file", Flags(OutputControl_Flag_CSV)),
      m_compression(m_aliases, "File compression", Flags(OutputControl_Flag_CSV | OutputControl_Flag_Generic)),
      m_binary(m_aliases, "File compression", Flags(OutputControl_Flag_CSV | OutputControl_Flag_Generic)),
      m_in_summary(m_aliases, "File in summary", Flags(OutputControl_Flag_CSV | OutputControl_Flag_Generic))
{
    //---------------------------------------
    m_enabled.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_Generic |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_file_enabled);

    m_enabled.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_enabled);

    //---------------------------------------
    m_pathes.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_Generic |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_file_path);

    m_pathes.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_path);

    //---------------------------------------
    m_csv_to_file.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_to_file);

    //---------------------------------------
    m_compression.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_Generic |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_file_compressed);

    m_compression.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_compressed);
    //---------------------------------------
    m_binary.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_binary);
    //---------------------------------------
    m_in_summary.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_Generic |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_file_in_summary);

    m_in_summary.set(
            Identity(
                Flags(
                    OutputControl_Flag_Valid |
                    OutputControl_Flag_CSV |
                    OutputControl_Flag_AppDefault)),
                app_settings.m_csv_in_summary);

}

std::ostream & OutputControl::output(std::ostream & stream, const std::string & prefix)
{
    stream << prefix << "Output control properties:" << std::endl;
    instance().enabled().output(stream, prefix + '\t');
    instance().pathes().output(stream, prefix + '\t');
    instance().csv_to_file().output(stream, prefix + '\t');
    instance().compression().output(stream, prefix + '\t');
    instance().binary().output(stream, prefix + '\t');

    return stream;
}

void OutputControl::Properties::init()
{
    std::string filename;

    if(!m_identity.is_valid())
        return;

    if(m_identity.is_csv())
    {
        if(!instance().csv_to_file().get(m_identity, m_csv_to_file).is_valid())
            return;

        if(!instance().binary().get(m_identity, m_binary).is_valid())
            return;

        filename = app_settings.m_csv_filename;
    }
    else
        filename = app_settings.m_file_filename;

    if(m_identity.flags() & OutputControl_Flag_UserFile)
    {
        if(m_identity.text().find_first_of('/') == std::string::npos) {
            if(!build_generic_path(m_identity.text()))
                return;
        }
        else
            m_path = m_identity.text();

        m_enabled  = true;
        m_is_valid = true;

        return;
    }
    else
    {
        if(!instance().enabled().get(m_identity, m_enabled).is_valid())
            return;

        const Identity & path_res = instance().pathes().get(m_identity, m_path);

        if(!path_res.is_valid())
            return;

        trim(m_path);

        if(m_identity.flags() & OutputControl_Flag_CustomFileName)
        {
            if(path_res.is_special())
            {
                if(*m_path.rbegin() != '/')
                    m_path += '/';
            }
            else
            {
            	rtrim(m_path, "/");

                std::string::size_type index = m_path.find_last_of('/');

                if(index != std::string::npos)
                    m_path += '/';
                else
                    if(!build_generic_path(m_path + '/'))
                        return;
            }

            std::string::size_type index = m_identity.text().find_last_of("/");

            m_path += index != std::string::npos ?
                                m_identity.text().substr(index + 1) :
                                m_identity.text();
        }
        else
            if(m_path.empty())
                return; // m_path = filename + '.' + m_identity.type();
            else
                if(path_res.is_special())
                {
                    if(*m_path.rbegin() == '/' || *m_path.rbegin() == '\\')
                        m_path += filename;

                    if(*m_path.rbegin() != '.')
                        m_path += '.';

                    m_path += m_identity.type();
                }
                else
                    // case when we have NOT special, but provided filename only
                    if(m_path.find_first_of('/') == std::string::npos)
                    {
                    	if(m_path.find_first_of('.') == std::string::npos)
                    		m_path += "." + m_identity.type();

                        if(!build_generic_path(m_path))
                            return;
                    }
                    else
                        // case when we have NOT special and filename is folder
                        if(*m_path.rbegin() == '/' || *m_path.rbegin() == '\\')
                            m_path += filename + '.' + m_identity.type();
    }

    if(!instance().compression().get(m_identity, m_compression).is_valid())
        return;

    if(!instance().in_summary().get(m_identity, m_in_summary).is_valid())
        return;

    m_is_valid = true;
}

bool OutputControl::Properties::build_generic_path(const std::string & path)
{
    std::string default_path;

    if(!instance().pathes().get(
        Flags(m_identity.flags() | OutputControl_Flag_Default), default_path).is_valid())
        return false;

    m_path = default_path + path;

    return true;
}

std::ostream & OutputControl::Properties::output(std::ostream & stream, const std::string & prefix)
{
    stream << prefix << "OutputControl::Properties:" << std::endl;
    m_identity.output(stream, prefix + "\t");
    stream << std::endl << prefix << "\tProperties: "
           << '[' << (m_is_valid ? "Valid" : "Invalid")
           << "|" << (m_enabled ? "Enabled" : "Disabled")
           << (m_compression ? ", |Compressed" : "");

            if(m_identity.is_csv())
            {
                stream << (m_binary ? "|Binary" : "")
                       << (m_csv_to_file ? "|CSV_TO_FILE" : "");
            }

    stream << ']' << " path is '" << m_path << "'";
    stream << std::endl;

    return stream;
}


OutputControl::AppSettings::AppSettings() {
    init("ibdiag_app");
}

//
//
//
#ifndef OUTPUT_CONTROL_BASE_PATH
    #define OUTPUT_CONTROL_BASE_PATH           "/var/tmp"
#endif


//
//
//
#ifndef OUTPUT_CONTROL_ENABLED
    #define OUTPUT_CONTROL_ENABLED              true
#endif

#ifndef OUTPUT_CONTROL_CSV_ENABLED
    #define OUTPUT_CONTROL_CSV_ENABLED          OUTPUT_CONTROL_ENABLED
#endif

#ifndef OUTPUT_CONTROL_FILES_ENABLED
    #define OUTPUT_CONTROL_FILE_ENABLED         OUTPUT_CONTROL_ENABLED
#endif

//
//
//
#ifndef OUTPUT_CONTROL_CSV_TO_FILE
    #define OUTPUT_CONTROL_CSV_TO_FILE          false
#endif

//
//
//
#ifndef OUTPUT_CONTROL_COMPRESSED
    #define OUTPUT_CONTROL_COMPRESSED           false
#endif

#ifndef OUTPUT_CONTROL_CSV_COMPRESSED
    #define OUTPUT_CONTROL_CSV_COMPRESSED       OUTPUT_CONTROL_COMPRESSED
#endif

#ifndef OUTPUT_CONTROL_FILE_COMPRESSED
    #define OUTPUT_CONTROL_FILE_COMPRESSED      OUTPUT_CONTROL_COMPRESSED
#endif

//
//
//
#ifndef OUTPUT_CONTROL_CSV_BINARY
    #define OUTPUT_CONTROL_CSV_BINARY           false
#endif

//
//
//
void OutputControl::AppSettings::init(const std::string & app_name)
{
    m_csv_enabled       = OUTPUT_CONTROL_CSV_ENABLED;
    m_file_enabled      = OUTPUT_CONTROL_FILE_ENABLED;

    m_csv_path          = std::string(OUTPUT_CONTROL_BASE_PATH) + "/" + app_name + "/";
    m_file_path         = std::string(OUTPUT_CONTROL_BASE_PATH) + "/" + app_name + "/";

    m_csv_to_file       = OUTPUT_CONTROL_CSV_TO_FILE;

    m_csv_compressed    = OUTPUT_CONTROL_CSV_COMPRESSED;
    m_file_compressed   = OUTPUT_CONTROL_FILE_COMPRESSED;

    m_csv_binary        = OUTPUT_CONTROL_CSV_BINARY;

    m_csv_in_summary    = true;
    m_file_in_summary   = true;

    m_csv_filename      = app_name + ".db_csv";
    m_file_filename     = app_name;
}

OutputControl::AppSettings OutputControl::app_settings;


bool OutputControl::CreateFolder(const std::string & full_path)
{
    if(full_path.empty())
        return 0;

    char buffer[1024];
         buffer[0] = full_path[0];

    for(unsigned int x = 1; x < full_path.size(); ++x)
    {
        char c = full_path[ x ];

        if(c == '/' && buffer[ x - 1 ] != c)
        {
            buffer[ x ] = '\0';

            if(mkdir(buffer, S_IRWXU | S_IRWXG | S_IRWXO) < 0)
                if(EEXIST != errno)
                    return false; // errno;
        }

        buffer[x] = c;
    }
    return true;
}

